# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from typing import Callable

from swh.counters import get_counters
from swh.counters.interface import CountersInterface
from swh.storage import StorageSpec, get_storage
from swh.storage.interface import StorageInterface

OBJECT_TYPES = [
    "content",
    "directory",
    "snapshot",
    "origin_visit_status",
    "origin_visit",
    "origin",
]


class CountingProxyStorage:
    """Counting Storage Proxy.

    This is in charge of adding objects directly to swh-counters, without
    going through Kafka/swh-journal.
    This is meant as a simple way to setup counters for experiments; production
    should use swh-journal to reduce load/latency of the storage server.

    Additionally, unlike the journal-based counting, it does not count persons
    or the number of origins per netloc.

    Sample configuration use case for filtering storage:

    .. code-block: yaml

        storage:
          cls: counter
          counters:
            cls: remote
            url: http://counters.internal.staging.swh.network:5011/
          storage:
            cls: remote
            url: http://storage.internal.staging.swh.network:5002/

    """

    def __init__(self, counters, storage: StorageSpec) -> None:
        self.counters: CountersInterface = get_counters(**counters)
        self.storage: StorageInterface = get_storage(**storage)

    def __getattr__(self, key):
        if key == "storage":
            raise AttributeError(key)
        if key.endswith("_add"):
            return self._adder(key[0:-4], getattr(self.storage, key))
        return getattr(self.storage, key)

    def _adder(self, collection: str, backend_function: Callable):
        def f(objs):
            self.counters.add(collection, [obj.unique_key() for obj in objs])
            return backend_function(objs)

        return f



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr
import pytest

from swh.storage import get_storage


@pytest.fixture
def swh_storage():
    storage_config = {
        "cls": "pipeline",
        "steps": [
            {"cls": "counter", "counters": {"cls": "memory"}},
            {"cls": "memory"},
        ],
    }

    return get_storage(**storage_config)


def test_counting_proxy_storage_content(swh_storage, sample_data):
    assert swh_storage.counters.counters["content"] == set()

    swh_storage.content_add([sample_data.content])

    assert swh_storage.counters.counters["content"] == {sample_data.content.sha1}

    swh_storage.content_add([sample_data.content2, sample_data.content3])

    assert swh_storage.counters.counters["content"] == {
        sample_data.content.sha1,
        sample_data.content2.sha1,
        sample_data.content3.sha1,
    }

    assert [
        attr.evolve(cnt, ctime=None)
        for cnt in swh_storage.content_find({"sha256": sample_data.content2.sha256})
    ] == [attr.evolve(sample_data.content2, data=None)]


def test_counting_proxy_storage_revision(swh_storage, sample_data):
    assert swh_storage.counters.counters["revision"] == set()

    swh_storage.revision_add([sample_data.revision])

    assert swh_storage.counters.counters["revision"] == {sample_data.revision.id}

    swh_storage.revision_add([sample_data.revision2, sample_data.revision3])

    assert swh_storage.counters.counters["revision"] == {
        sample_data.revision.id,
        sample_data.revision2.id,
        sample_data.revision3.id,
    }

    assert swh_storage.revision_get([sample_data.revision2.id]) == [
        sample_data.revision2
    ]



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import Mock

import attr
import pytest

from swh.storage import get_storage


@pytest.fixture
def swh_storage():
    storage_config = {
        "cls": "pipeline",
        "steps": [
            {"cls": "filter"},
            {"cls": "memory"},
        ],
    }

    return get_storage(**storage_config)


def test_filtering_proxy_storage_content(swh_storage, sample_data):
    sample_content = sample_data.content

    content = swh_storage.content_get_data(sample_content.sha1)
    assert content is None

    s = swh_storage.content_add([sample_content])
    assert s == {
        "content:add": 1,
        "content:add:bytes": sample_content.length,
    }

    content = swh_storage.content_get_data(sample_content.sha1)
    assert content is not None

    s = swh_storage.content_add([sample_content])
    assert s == {
        "content:add": 0,
        "content:add:bytes": 0,
    }


def test_filtering_proxy_storage_skipped_content(swh_storage, sample_data):
    sample_content = sample_data.skipped_content
    sample_content_dict = sample_content.to_dict()

    content = next(swh_storage.skipped_content_missing([sample_content_dict]))
    assert content["sha1"] == sample_content.sha1

    s = swh_storage.skipped_content_add([sample_content])
    assert s == {
        "skipped_content:add": 1,
    }

    content = list(swh_storage.skipped_content_missing([sample_content_dict]))
    assert content == []

    s = swh_storage.skipped_content_add([sample_content])
    assert s == {
        "skipped_content:add": 0,
    }


def test_filtering_proxy_storage_skipped_content_missing_sha1_git(
    swh_storage, sample_data
):
    sample_contents = [
        attr.evolve(c, sha1_git=None) for c in sample_data.skipped_contents
    ]
    sample_content, sample_content2 = [c.to_dict() for c in sample_contents[:2]]

    content = next(swh_storage.skipped_content_missing([sample_content]))
    assert content["sha1"] == sample_content["sha1"]

    s = swh_storage.skipped_content_add([sample_contents[0]])
    assert s == {
        "skipped_content:add": 1,
    }

    content = list(swh_storage.skipped_content_missing([sample_content]))
    assert content == []

    s = swh_storage.skipped_content_add([sample_contents[1]])
    assert s == {
        "skipped_content:add": 1,
    }

    content = list(swh_storage.skipped_content_missing([sample_content2]))
    assert content == []


def test_filtering_proxy_storage_revision(swh_storage, sample_data):
    sample_revision = sample_data.revision

    revision = swh_storage.revision_get([sample_revision.id])[0]
    assert revision is None

    s = swh_storage.revision_add([sample_revision])
    assert s == {
        "revision:add": 1,
    }

    revision = swh_storage.revision_get([sample_revision.id])[0]
    assert revision is not None

    s = swh_storage.revision_add([sample_revision])
    assert s == {
        "revision:add": 0,
    }


def test_filtering_proxy_storage_release(swh_storage, sample_data):
    sample_release = sample_data.release

    release = swh_storage.release_get([sample_release.id])[0]
    assert release is None

    s = swh_storage.release_add([sample_release])
    assert s == {
        "release:add": 1,
    }

    release = swh_storage.release_get([sample_release.id])[0]
    assert release is not None

    s = swh_storage.release_add([sample_release])
    assert s == {
        "release:add": 0,
    }


def test_filtering_proxy_storage_directory(swh_storage, sample_data):
    sample_directory = sample_data.directory

    directory = list(swh_storage.directory_missing([sample_directory.id]))[0]
    assert directory

    s = swh_storage.directory_add([sample_directory])
    assert s == {
        "directory:add": 1,
    }

    directory = list(swh_storage.directory_missing([sample_directory.id]))
    assert not directory

    s = swh_storage.directory_add([sample_directory])
    assert s == {
        "directory:add": 0,
    }


def test_filtering_proxy_storage_empty_list(swh_storage, sample_data):
    swh_storage.storage = mock_storage = Mock(wraps=swh_storage.storage)

    calls = 0
    for object_type in swh_storage.object_types:
        calls += 1

        method_name = f"{object_type}_add"
        method = getattr(swh_storage, method_name)
        one_object = getattr(sample_data, object_type)

        # Call with empty list: ensure underlying storage not called
        method([])
        assert method_name not in {c[0] for c in mock_storage.method_calls}
        mock_storage.reset_mock()

        # Call with an object: ensure underlying storage is called
        method([one_object])
        assert method_name in {c[0] for c in mock_storage.method_calls}
        mock_storage.reset_mock()

        # Call with the same object: ensure underlying storage is not called again
        method([one_object])
        assert method_name not in {c[0] for c in mock_storage.method_calls}
        mock_storage.reset_mock()

    assert calls > 0, "Empty list never tested"



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

import pytest

from swh.storage.metrics import (
    OPERATIONS_METRIC,
    OPERATIONS_UNIT_METRIC,
    DifferentialTimer,
    send_metric,
)


def test_send_metric_unknown_unit():
    r = send_metric("content", count=10, method_name="content_add")
    assert r is False
    r = send_metric("sthg:add:bytes:extra", count=10, method_name="sthg_add")
    assert r is False


def test_send_metric_no_value():
    r = send_metric("content:add", count=0, method_name="content_add")
    assert r is False


@patch("swh.storage.metrics.statsd.increment")
def test_send_metric_no_unit(mock_statsd):
    r = send_metric("content:add", count=10, method_name="content_add")

    mock_statsd.assert_called_with(
        OPERATIONS_METRIC,
        10,
        tags={
            "endpoint": "content_add",
            "object_type": "content",
            "operation": "add",
        },
    )

    assert r


@patch("swh.storage.metrics.statsd.increment")
def test_send_metric_unit(mock_statsd):
    unit_ = "bytes"
    r = send_metric("c:add:%s" % unit_, count=100, method_name="c_add")

    expected_metric = OPERATIONS_UNIT_METRIC.format(unit=unit_)
    mock_statsd.assert_called_with(
        expected_metric,
        100,
        tags={
            "endpoint": "c_add",
            "object_type": "c",
            "operation": "add",
        },
    )

    assert r


def test_differential_timer(mocker):
    mocked_timing = mocker.patch("swh.core.statsd.statsd.timing")
    mocked_monotonic = mocker.patch("swh.storage.metrics.monotonic")
    mocked_monotonic.side_effect = [
        0.0,  # outer start
        1.0,  # inner.start 1
        1.5,  # inner.done 1 (0.5s elapsed)
        1.7,  # inner.start 2
        1.9,  # inner.done 2 (0.2s elapsed)
        4.5,  # outer done (4.5s elapsed)
        Exception("Should not have been called again!"),
    ]

    timer = DifferentialTimer("metric_difftimer_seconds", tags={"test": "value"})

    def do_something():
        pass

    def do_something_else():
        pass

    with timer:
        with timer.inner():
            do_something()
        with timer.inner():
            do_something()
        do_something_else()

    mocked_timing.assert_called_once_with(
        "metric_difftimer_seconds",
        # The differential timing is 3800 ms: 4.5s - 0.5s - 0.2s
        pytest.approx(3800.0),
        tags={"test": "value"},
    )


def test_differential_timer_inner_exception(mocker):
    mocked_timing = mocker.patch("swh.core.statsd.statsd.timing")
    mocked_monotonic = mocker.patch("swh.storage.metrics.monotonic")
    mocked_monotonic.side_effect = [
        5.0,  # outer start
        6.0,  # inner start
        # exception raised by inner
        6.1,  # inner finished (0.1s elapsed)
        6.2,  # outer finished (1.2s elapsed)
        Exception("Should not have been called again!"),
    ]

    timer = DifferentialTimer("metric_innerexception_seconds", tags={"test2": "value"})

    class InnerException(Exception):
        pass

    with pytest.raises(InnerException):
        with timer:
            with timer.inner():
                raise InnerException

            assert False, "unreachable"

    mocked_timing.assert_called_once_with(
        "metric_innerexception_seconds",
        # The differential timing is 1100 ms: 1.2s - 0.1s
        pytest.approx(1100.0),
        tags={"test2": "value", "inner_exc": "InnerException"},
    )


def test_differential_timer_outer_exception(mocker):
    mocked_timing = mocker.patch("swh.core.statsd.statsd.timing")
    mocked_monotonic = mocker.patch("swh.storage.metrics.monotonic")
    mocked_monotonic.side_effect = [
        15.0,  # outer start
        17.0,  # inner start
        17.7,  # inner finished (0.7s elapsed)
        # exception raised by outer
        18.2,  # outer finished (3.2s elapsed)
        Exception("Should not have been called again!"),
    ]

    timer = DifferentialTimer("metric_outerexception_seconds")

    class OuterException(Exception):
        pass

    with pytest.raises(OuterException):
        with timer:
            with timer.inner():
                pass
            raise OuterException

            assert False, "unreachable"

    mocked_timing.assert_called_once_with(
        "metric_outerexception_seconds",
        # The differential timing is 2500 ms: 3.2s - 0.7s
        pytest.approx(2500.0),
        tags={"outer_exc": "OuterException"},
    )



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict, List

from swh.core.api import RemoteException, RPCClient
from swh.model.model import Content

from ..exc import (
    BlockedOriginException,
    HashCollision,
    MaskedObjectException,
    NonRetryableException,
    QueryTimeout,
    StorageAPIError,
    StorageArgumentException,
    UnknownMetadataAuthority,
    UnknownMetadataFetcher,
)
from ..interface import StorageInterface
from .serializers import DECODERS, ENCODERS


class RemoteStorage(RPCClient):
    """Proxy to a remote storage API"""

    api_exception = StorageAPIError
    backend_class = StorageInterface
    reraise_exceptions = [
        BlockedOriginException,
        MaskedObjectException,
        NonRetryableException,
        QueryTimeout,
        StorageArgumentException,
        UnknownMetadataAuthority,
        UnknownMetadataFetcher,
    ]
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS

    def raise_for_status(self, response) -> None:
        try:
            super().raise_for_status(response)
        except RemoteException as e:
            if (
                e.response is not None
                and e.response.status_code == 500
                and e.args
                and e.args[0].get("type") == "HashCollision"
            ):
                # XXX: workaround until we fix these HashCollisions happening
                # when they shouldn't
                raise HashCollision(*e.args[0]["args"])
            else:
                raise

    def content_add(self, content: List[Content]) -> Dict[str, int]:
        return self._post("content/add", {"content": [c.with_data() for c in content]})

    def reset(self):
        return self._post("reset", {})

    def stat_counters(self):
        return self.get("stat/counters")

    def refresh_stat_counters(self):
        return self.get("stat/refresh")



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from itertools import chain
from typing import Iterable

import pytest

from swh.model.model import Origin, OriginVisit, OriginVisitStatus
from swh.storage.exc import BlockedOriginException
from swh.storage.proxies.blocking import BlockingProxyStorage
from swh.storage.proxies.blocking.db import BlockingState


def now() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


@pytest.fixture
def swh_storage_backend_config():
    return {
        "cls": "memory",
        "journal_writer": {
            "cls": "memory",
        },
    }


@pytest.fixture
def swh_storage(blocking_db_postgresql, swh_storage_backend):
    return BlockingProxyStorage(
        blocking_db=blocking_db_postgresql.info.dsn, storage=swh_storage_backend
    )


def set_origin_visibility(blocking_admin, slug="foo", reason="bar"):
    # Create a request
    request = blocking_admin.create_request(slug=slug, reason=reason)

    def set_visibility(urls: Iterable[str], new_state: BlockingState):
        blocking_admin.set_origins_state(
            request_id=request.id,
            new_state=new_state,
            urls=list(urls),
        )

    return request, set_visibility


def mk_origin_visit(origin: Origin):
    return OriginVisit(origin=origin.url, date=now(), type="git")


def mk_origin_visit_status(origin: Origin):
    return OriginVisitStatus(
        origin=origin.url, visit=1, date=now(), status="created", snapshot=None
    )


OBJ_FACTORY = {
    "origin_add": lambda x: x,
    "origin_visit_add": mk_origin_visit,
    "origin_visit_status_add": mk_origin_visit_status,
}


@pytest.mark.parametrize(
    "endpoint", ["origin_add", "origin_visit_add", "origin_visit_status_add"]
)
def test_blocking_prefix_match_simple(swh_storage, blocking_admin, endpoint):
    request, set_visibility = set_origin_visibility(blocking_admin)
    method = getattr(swh_storage, endpoint)
    obj_factory = OBJ_FACTORY[endpoint]
    backend_storage = swh_storage.storage

    # beware of the mix of example.com vs example.org URLs below
    url_patterns = {
        BlockingState.DECISION_PENDING: ["https://example.com/user1/repo1"],
        BlockingState.BLOCKED: ["https://example.org/user1"],
    }
    for decision, urls in url_patterns.items():
        set_visibility(urls, decision)

    origins = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user1/repo1/",
            "https://example.com/user1/repo1.git",
            "https://example.com/user1/repo1.git/",
            "https://example.com/user1/repo1/subrepo1",
            "https://example.com/user1/repo1/subrepo1.git",
        ],
        BlockingState.BLOCKED: [
            "https://example.org/user1",
            "https://example.org/user1/repo1",
        ],
        BlockingState.NON_BLOCKED: [
            "https://example.com/user1",
            "https://example.com/user1/repo2",
            "https://example.com/user1/repo2/",
            "https://example.com/user1/repo2.git",
            "https://example.com/user1/repo2.git/",
            "https://example.com/user1/repo2/subrepo1",
            "https://example.com/user1/repo2/subrepo1.git",
            "https://example.com/user2/repo1",
            "https://example.com/user2/repo1",
            "https://example.com/user11",
            "https://example.com/user11/repo1",
            "https://example.org/user11",
            "https://example.org/user11/repo1",
        ],
    }
    all_origins = [Origin(url) for url in chain(*(origins.values()))]

    if endpoint != "origin_add":
        # if the endpoint is not 'origin_add', actually insert the origins; no
        # visit or visit status should be able to be inserted nonetheless
        assert backend_storage.origin_add(all_origins)
    if endpoint == "origin_visit_status_add":
        # if the endpoint is 'origin_visit_status_add', actually insert the
        # origin visits in the backend; but no visit status should be able to
        # be inserted via the proxy
        assert backend_storage.origin_visit_add(
            [mk_origin_visit(origin) for origin in all_origins]
        )

    for state, urls in origins.items():
        if state != BlockingState.NON_BLOCKED:
            for url in urls:
                origin = Origin(url=url)
                with pytest.raises(BlockedOriginException) as exc_info:
                    method([obj_factory(origin)])
                    print("URL should not be inserted", origin.url)
                assert {
                    url: status.state for url, status in exc_info.value.blocked.items()
                } == {origin.url: state}
        else:
            # this should not be blocked
            for url in urls:
                origin = Origin(url=url)
                assert method([obj_factory(origin)])


@pytest.mark.parametrize(
    "endpoint", ["origin_add", "origin_visit_add", "origin_visit_status_add"]
)
def test_blocking_prefix_match_with_overload(swh_storage, blocking_admin, endpoint):
    request, set_visibility = set_origin_visibility(blocking_admin)
    method = getattr(swh_storage, endpoint)
    obj_factory = OBJ_FACTORY[endpoint]
    backend_storage = swh_storage.storage

    url_patterns = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user2",
        ],
        BlockingState.NON_BLOCKED: [
            "https://example.com/user2/repo1",
        ],
        BlockingState.BLOCKED: [
            "https://example.com/user2/repo1/subrepo1",
        ],
    }

    for decision, urls in url_patterns.items():
        set_visibility(urls, decision)

    origins = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user1/repo1/",
            "https://example.com/user1/repo1.git",
            "https://example.com/user1/repo1.git/",
            "https://example.com/user1/repo1/subrepo1",
            "https://example.com/user1/repo1/subrepo1.git",
            # everything under https://example.com/user2 should be blocked (but
            # https://example.com/user2/repo1)
            "https://example.com/user2",
            "https://example.com/user2.git",
            "https://example.com/user2.git/",
            "https://example.com/user2/",
            "https://example.com/user2/repo2",
            "https://example.com/user2/repo2/",
            "https://example.com/user2/repo2.git",
            "https://example.com/user2/repo2.git/",
        ],
        BlockingState.BLOCKED: [
            # "https://example.com/user2/repo1/subrepo1 is explicitly blocked
            "https://example.com/user2/repo1/subrepo1",
            "https://example.com/user2/repo1/subrepo1.git",
            "https://example.com/user2/repo1/subrepo1/subsubrepo",
        ],
        BlockingState.NON_BLOCKED: [
            "https://example.com/user1",
            "https://example.com/user1/repo2",
            # https://example.com/user2/repo1 is explicitly enabled
            "https://example.com/user2/repo1",
            "https://example.com/user2/repo1/",
            "https://example.com/user2/repo1.git",
            "https://example.com/user2/repo1/subrepo2",
        ],
    }
    all_origins = [Origin(url) for url in chain(*(origins.values()))]

    if endpoint != "origin_add":
        # if the endpoint is not 'origin_add', actually insert the origins; no
        # visit or visit status should be able to be inserted nonetheless
        assert backend_storage.origin_add(all_origins)
    if endpoint == "origin_visit_status_add":
        # if the endpoint is 'origin_visit_status_add', actually insert the
        # origin visits in the backend; but no visit status should be able to
        # be inserted via the proxy
        assert backend_storage.origin_visit_add(
            [mk_origin_visit(origin) for origin in all_origins]
        )

    for state, urls in origins.items():
        if state != BlockingState.NON_BLOCKED:
            for url in urls:
                origin = Origin(url=url)
                with pytest.raises(BlockedOriginException) as exc_info:
                    method([obj_factory(origin)])
                    print("URL should not be inserted", origin.url)
                assert {
                    url: status.state for url, status in exc_info.value.blocked.items()
                } == {origin.url: state}
        else:
            # this should not be blocked
            for url in urls:
                origin = Origin(url=url)
                assert method([obj_factory(origin)])


@pytest.mark.parametrize(
    "endpoint", ["origin_add", "origin_visit_add", "origin_visit_status_add"]
)
def test_blocking_prefix_match_add_multi(swh_storage, blocking_admin, endpoint):
    request, set_visibility = set_origin_visibility(blocking_admin)
    method = getattr(swh_storage, endpoint)
    obj_factory = OBJ_FACTORY[endpoint]
    backend_storage = swh_storage.storage

    url_patterns = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user2",
        ],
        BlockingState.NON_BLOCKED: [
            "https://example.com/user2/repo1",
        ],
        BlockingState.BLOCKED: [
            "https://example.com/user2/repo1/subrepo1",
        ],
    }

    requests = {}
    set_vs = {}
    for decision, urls in url_patterns.items():
        request, set_visibility = set_origin_visibility(
            blocking_admin, slug=f"foo_{decision.name}"
        )
        set_visibility(urls, decision)
        requests[decision] = request
        set_vs[decision] = set_visibility

    origins = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user1/repo1/",
            "https://example.com/user1/repo1.git",
            "https://example.com/user1/repo1.git/",
            "https://example.com/user1/repo1/subrepo1",
            "https://example.com/user1/repo1/subrepo1.git",
            # everything under https://example.com/user2 should be blocked (but
            # https://example.com/user2/repo1)
            "https://example.com/user2",
            "https://example.com/user2.git",
            "https://example.com/user2.git/",
            "https://example.com/user2/",
            "https://example.com/user2/repo2",
            "https://example.com/user2/repo2/",
            "https://example.com/user2/repo2.git",
            "https://example.com/user2/repo2.git/",
        ],
        BlockingState.NON_BLOCKED: [
            # these should work OK by themselves
            "https://example.com/user1",
            "https://example.com/user1/repo2",
            # https://example.com/user2/repo1 is explicitly enabled
            "https://example.com/user2/repo1",
            "https://example.com/user2/repo1/",
            "https://example.com/user2/repo1.git",
            "https://example.com/user2/repo1/subrepo2",
        ],
        BlockingState.BLOCKED: [
            # "https://example.com/user2/repo1/subrepo1 is explicitly blocked
            "https://example.com/user2/repo1/subrepo1",
            "https://example.com/user2/repo1/subrepo1.git",
            "https://example.com/user2/repo1/subrepo1/subsubrepo",
        ],
    }

    # insert them all at once, none should be inserted and the error give
    # reasons for rejected urls
    all_origins = [Origin(url) for url in chain(*(origins.values()))]

    if endpoint != "origin_add":
        # if the endpoint is not 'origin_add', insert the origins; no
        # visit or visit status should be able to be inserted nonetheless
        assert backend_storage.origin_add(all_origins)
    if endpoint == "origin_visit_status_add":
        # if the endpoint is 'origin_visit_status_add', insert the
        # origin visits in the backend; but no visit status should be able to
        # be inserted via the proxy
        assert backend_storage.origin_visit_add(
            [mk_origin_visit(origin) for origin in all_origins]
        )
    # the insertion of all the objects at once fails because there are blocked
    # elements in the batch
    with pytest.raises(BlockedOriginException) as exc_info:
        method([obj_factory(origin) for origin in all_origins])
    # check the reported list of actually blocked origins is ok
    assert {
        url
        for url, status in exc_info.value.blocked.items()
        if status.state == BlockingState.DECISION_PENDING
    } == set(origins[BlockingState.DECISION_PENDING])
    assert {
        url
        for url, status in exc_info.value.blocked.items()
        if status.state == BlockingState.BLOCKED
    } == set(origins[BlockingState.BLOCKED])

    # now, set the decision pending pattern to non_blocked and rerun the whole insertion thing
    # use the set_visibility helper coming the the request that created decision pending rules
    set_visibility = set_vs[BlockingState.DECISION_PENDING]
    set_visibility(
        url_patterns[BlockingState.DECISION_PENDING], BlockingState.NON_BLOCKED
    )

    with pytest.raises(BlockedOriginException) as exc_info:
        method([obj_factory(origin) for origin in all_origins])
    # urls from the (formerly) decision pending batch should now not be reported
    # as blocked in the exception
    assert not {
        url
        for url, status in exc_info.value.blocked.items()
        if status.state == BlockingState.DECISION_PENDING
    }
    assert {
        url
        for url, status in exc_info.value.blocked.items()
        if status.state == BlockingState.BLOCKED
    } == set(origins[BlockingState.BLOCKED])

    # now if we ingest only non blocked urls, it should be ok (aka all but the
    # urls in origins[BLOCKED])
    assert method(
        [
            obj_factory(origin)
            for origin in all_origins
            if origin.url not in origins[BlockingState.BLOCKED]
        ]
    )


@pytest.mark.parametrize(
    "endpoint", ["origin_add", "origin_visit_add", "origin_visit_status_add"]
)
def test_blocking_log(swh_storage, blocking_admin, endpoint):
    method = getattr(swh_storage, endpoint)
    obj_factory = OBJ_FACTORY[endpoint]
    backend_storage = swh_storage.storage

    url_patterns = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user2",
        ],
        BlockingState.NON_BLOCKED: [
            "https://example.com/user2/repo1",
        ],
        BlockingState.BLOCKED: [
            "https://example.com/user2/repo1/subrepo1",
        ],
    }
    requests = {}
    set_vs = {}
    for decision, urls in url_patterns.items():
        request, set_visibility = set_origin_visibility(
            blocking_admin, slug=f"foo_{decision.name}"
        )
        set_visibility(urls, decision)
        requests[decision] = request
        set_vs[decision] = set_visibility

    origins = {
        BlockingState.DECISION_PENDING: [
            "https://example.com/user1/repo1",
            "https://example.com/user1/repo1/",
            "https://example.com/user1/repo1.git",
            "https://example.com/user1/repo1.git/",
            "https://example.com/user1/repo1/subrepo1",
            "https://example.com/user1/repo1/subrepo1.git",
            # everything under https://example.com/user2 should be blocked (but
            # https://example.com/user2/repo1)
            "https://example.com/user2",
            "https://example.com/user2.git",
            "https://example.com/user2.git/",
            "https://example.com/user2/",
            "https://example.com/user2/repo2",
            "https://example.com/user2/repo2/",
            "https://example.com/user2/repo2.git",
            "https://example.com/user2/repo2.git/",
        ],
        None: [
            # these should work OK by themselves, not matching any rule
            "https://example.com/user1",
            "https://example.com/user1/repo2",
            "https://example.org/user2/repo2",
            "https://example.org/user3/repo2/",
        ],
        BlockingState.NON_BLOCKED: [
            # https://example.com/user2/repo1 is explicitly enabled
            "https://example.com/user2/repo1",
            "https://example.com/user2/repo1/",
            "https://example.com/user2/repo1.git",
            "https://example.com/user2/repo1/subrepo2",
        ],
        BlockingState.BLOCKED: [
            # "https://example.com/user2/repo1/subrepo1 is explicitly blocked
            "https://example.com/user2/repo1/subrepo1",
            "https://example.com/user2/repo1/subrepo1.git",
            "https://example.com/user2/repo1/subrepo1/subsubrepo",
        ],
    }

    # insert them all at once, none should be inserted and the error give
    # reasons for rejected urls
    all_origins = [Origin(url) for url in chain(*(origins.values()))]

    ts_before = now()
    if endpoint != "origin_add":
        # if the endpoint is not 'origin_add', insert the origins; no
        # visit or visit status should be able to be inserted nonetheless
        assert backend_storage.origin_add(all_origins)
    if endpoint == "origin_visit_status_add":
        # if the endpoint is 'origin_visit_status_add', insert the
        # origin visits in the backend; but no visit status should be able to
        # be inserted via the proxy
        assert backend_storage.origin_visit_add(
            [mk_origin_visit(origin) for origin in all_origins]
        )

    with pytest.raises(BlockedOriginException):
        method([obj_factory(origin) for origin in all_origins])
    ts_after = now()

    # Check blocking journal log
    log = blocking_admin.get_log()
    assert set(origins[BlockingState.DECISION_PENDING]) == set(
        x.url for x in log if x.state == BlockingState.DECISION_PENDING
    )
    assert set(origins[BlockingState.BLOCKED]) == set(
        x.url for x in log if x.state == BlockingState.BLOCKED
    )
    assert set(origins[BlockingState.NON_BLOCKED]) == set(
        x.url for x in log if x.state == BlockingState.NON_BLOCKED
    )
    urls_with_event = {x.url for x in log}
    for url in origins[None]:
        assert url not in urls_with_event

    for logentry in log:
        assert logentry.request == requests[logentry.state].id
        assert ts_before < logentry.date < ts_after

    # check the get_log(request_id=xx) method call
    for state, request in requests.items():
        log = blocking_admin.get_log(request_id=request.id)
        assert set(origins[state]) == set(x.url for x in log)

    # check the get_log(url=xx) method call
    for origin in all_origins:
        url = origin.url
        log = blocking_admin.get_log(url=url)
        if url in origins[None]:
            assert not log
        else:
            assert len(log) == 1
            logentry = log[0]
            assert logentry.url == url
            assert logentry.request == requests[logentry.state].id

    # now, set the decision pending pattern to non_blocked and rerun the whole
    # insertion thing; use the set_visibility helper coming with the request
    # that created decision pending rules
    set_visibility = set_vs[BlockingState.DECISION_PENDING]
    set_visibility(
        url_patterns[BlockingState.DECISION_PENDING], BlockingState.NON_BLOCKED
    )

    with pytest.raises(BlockedOriginException):
        method([obj_factory(origin) for origin in all_origins])

    # for each url in origin[BLOCKED], we should have 2 'blocked' log entries
    for url in origins[BlockingState.BLOCKED]:
        log = blocking_admin.get_log(url=url)
        assert len(log) == 2
        assert {entry.state for entry in log} == {BlockingState.BLOCKED}

    # for each url in origin[NON_BLOCKED], we should have 2 'non-blocked' log entries
    for url in origins[BlockingState.NON_BLOCKED]:
        log = blocking_admin.get_log(url=url)
        assert len(log) == 2
        assert {entry.state for entry in log} == {BlockingState.NON_BLOCKED}

    # for each url in origin[DESCISION_PENDING], we should have 1 'decision-pending' and
    # 1 'non-blocked' log entry
    for url in origins[BlockingState.DECISION_PENDING]:
        log = blocking_admin.get_log(url=url)
        assert len(log) == 2
        assert {entry.state for entry in log} == {
            BlockingState.NON_BLOCKED,
            BlockingState.DECISION_PENDING,
        }

    # for each url in origin[NON_BLOCKED], we should have 2 'non-blocked' log entries
    for url in origins[BlockingState.NON_BLOCKED]:
        log = blocking_admin.get_log(url=url)
        assert len(log) == 2
        assert {entry.state for entry in log} == {BlockingState.NON_BLOCKED}

    # for each url in origin[None], we should have no log entry
    for url in origins[None]:
        log = blocking_admin.get_log(url=url)
        assert len(log) == 0


def test_proxy_config_deprecation(blocking_db_postgresql, swh_storage_backend):
    with pytest.warns(DeprecationWarning):
        assert BlockingProxyStorage(
            blocking_db=blocking_db_postgresql.info.dsn, storage=swh_storage_backend
        )



# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, TypeVar, Union

import attr
from typing_extensions import Protocol, TypedDict, runtime_checkable

from swh.core.api import remote_api_endpoint
from swh.core.api.classes import PagedResult as CorePagedResult
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ExtID,
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    RawExtrinsicMetadata,
    Release,
    Revision,
    Sha1,
    Sha1Git,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
)
from swh.model.swhids import CoreSWHID, ExtendedSWHID, ObjectType


class ListOrder(Enum):
    """Specifies the order for paginated endpoints returning sorted results."""

    ASC = "asc"
    DESC = "desc"


class PartialBranches(TypedDict):
    """Type of the dictionary returned by snapshot_get_branches"""

    id: Sha1Git
    """Identifier of the snapshot"""
    branches: Dict[bytes, Optional[SnapshotBranch]]
    """A dict of branches contained in the snapshot
    whose keys are the branches' names"""
    next_branch: Optional[bytes]
    """The name of the first branch not returned or :const:`None` if
    the snapshot has less than the request number of branches."""


@attr.s
class SnapshotBranchByNameResponse:
    """Object returned by snapshot_branch_get_by_name"""

    branch_found = attr.ib(type=bool)
    """
    Branch with the name exists, with or without a target.
    """
    target = attr.ib(type=Optional[SnapshotBranch])
    """
    Branch target, will be None in case of a dangling branch.
    """
    aliases_followed = attr.ib(type=List[bytes])
    """
    List of alias names until (including) the target.
    This will be of length one for all non alias branches.
    """


class HashDict(TypedDict, total=False):
    sha1: bytes
    sha1_git: bytes
    sha256: bytes
    blake2s256: bytes


class TotalHashDict(HashDict, total=True):
    pass


@attr.s
class OriginVisitWithStatuses:
    visit = attr.ib(type=OriginVisit)
    statuses = attr.ib(type=List[OriginVisitStatus])


@attr.s
class ObjectReference:
    """Record that the object with SWHID ``source`` references the object with SWHID
    ``target``, meaning that the ``target`` needs to exist for the ``source`` object
    to be consistent within the archive."""

    source = attr.ib(type=ExtendedSWHID)
    target = attr.ib(type=ExtendedSWHID)


TResult = TypeVar("TResult")
PagedResult = CorePagedResult[TResult, str]


# TODO: Make it an enum (too much impact)
VISIT_STATUSES = ["created", "ongoing", "full", "partial"]


def deprecated(f):
    f.deprecated_endpoint = True
    return f


@runtime_checkable
class StorageInterface(Protocol):
    @remote_api_endpoint("check_config")
    def check_config(self, *, check_write: bool) -> bool:
        """Check that the storage is configured and ready to go."""
        ...

    ##########################
    # Content
    ##########################

    @remote_api_endpoint("content/add")
    def content_add(self, content: List[Content]) -> Dict[str, int]:
        """Add content blobs to the storage

        Args:
            contents (iterable): iterable of dictionaries representing
                individual pieces of content to add. Each dictionary has the
                following keys:

                - data (bytes): the actual content
                - length (int): content length
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.ALGORITHMS`, mapped to the
                  corresponding checksum
                - status (str): one of visible, hidden

        Raises:

            The following exceptions can occur:

            - HashCollision in case of collision
            - Any other exceptions raise by the db

            In case of errors, some of the content may have been stored in
            the DB and in the objstorage.
            Since additions to both idempotent, that should not be a problem.

        Returns:
            Summary dict with the following keys and associated values:

                content:add: New contents added
                content:add:bytes: Sum of the contents' length data
        """
        ...

    @remote_api_endpoint("content/update")
    def content_update(
        self, contents: List[Dict[str, Any]], keys: List[str] = []
    ) -> None:
        """Update content blobs to the storage. Does nothing for unknown
        contents or skipped ones.

        Args:
            content: iterable of dictionaries representing
                individual pieces of content to update. Each dictionary has the
                following keys:

                - data (bytes): the actual content
                - length (int): content length (default: -1)
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.ALGORITHMS`, mapped to the
                  corresponding checksum
                - status (str): one of visible, hidden, absent

            keys (list): List of keys (str) whose values needs an update, e.g.,
                new hash column

        """
        ...

    @remote_api_endpoint("content/add_metadata")
    def content_add_metadata(self, content: List[Content]) -> Dict[str, int]:
        """Add content metadata to the storage (like `content_add`, but
        without inserting to the objstorage).

        Args:
            content (iterable): iterable of dictionaries representing
                individual pieces of content to add. Each dictionary has the
                following keys:

                - length (int): content length (default: -1)
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.ALGORITHMS`, mapped to the
                  corresponding checksum
                - status (str): one of visible, hidden, absent
                - reason (str): if status = absent, the reason why
                - origin (int): if status = absent, the origin we saw the
                  content in
                - ctime (datetime): time of insertion in the archive

        Returns:
            Summary dict with the following key and associated values:

                content:add: New contents added
                skipped_content:add: New skipped contents (no data) added
        """
        ...

    @remote_api_endpoint("content/data")
    def content_get_data(self, content: Union[HashDict, Sha1]) -> Optional[bytes]:
        """Given a content identifier, returns its associated data if any.

        Args:
            content: dict of hashes (or just sha1 identifier)

        Returns:
             raw content data (bytes)

        """
        ...

    @remote_api_endpoint("content/partition")
    def content_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Content]:
        """Splits contents into nb_partitions, and returns one of these based on
        partition_id (which must be in [0, nb_partitions-1])

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id: index of the partition to fetch
            nb_partitions: total number of partitions to split into
            page_token: opaque token used for pagination.
            limit: Limit result (default to 1000)

        Returns:
            PagedResult of Content model objects within the partition. If
            next_page_token is None, there is no longer data to retrieve.

        """
        ...

    @remote_api_endpoint("content/metadata")
    def content_get(
        self, contents: List[bytes], algo: str = "sha1"
    ) -> List[Optional[Content]]:
        """Retrieve content metadata in bulk

        Args:
            content: List of content identifiers
            algo: one of the checksum algorithm in
              :data:`swh.model.hashutil.DEFAULT_ALGORITHMS`

        Returns:
            List of contents model objects when they exist, None otherwise.

        """
        ...

    @remote_api_endpoint("content/missing")
    def content_missing(
        self, contents: List[HashDict], key_hash: str = "sha1"
    ) -> Iterable[bytes]:
        """List content missing from storage

        Args:
            content: iterable of dictionaries whose keys are either 'length' or an item
                of :data:`swh.model.hashutil.ALGORITHMS`; mapped to the
                corresponding checksum (or length).
            key_hash: name of the column to use as hash id result (default: 'sha1')

        Raises:
            StorageArgumentException when key_hash is unknown.
            TODO: an exception when we get a hash collision.

        Returns:
            iterable of missing content ids (as per the `key_hash` column)

        """
        ...

    @remote_api_endpoint("content/missing/sha1")
    def content_missing_per_sha1(self, contents: List[bytes]) -> Iterable[bytes]:
        """List content missing from storage based only on sha1.

        Args:
            contents: List of sha1 to check for absence.

        Raises:
            TODO: an exception when we get a hash collision.

        Returns:
            Iterable of missing content ids (sha1)

        """
        ...

    @remote_api_endpoint("content/missing/sha1_git")
    def content_missing_per_sha1_git(
        self, contents: List[Sha1Git]
    ) -> Iterable[Sha1Git]:
        """List content missing from storage based only on sha1_git.

        Args:
            contents (List): An iterable of content id (sha1_git)

        Yields:
            missing contents sha1_git

        """
        ...

    @remote_api_endpoint("content/present")
    def content_find(self, content: HashDict) -> List[Content]:
        """Find a content hash in db.

        Args:
            content: a dictionary representing one content hash, mapping
                checksum algorithm names (see swh.model.hashutil.ALGORITHMS) to
                checksum values

        Raises:
            ValueError: in case the key of the dictionary is not sha1, sha1_git
                nor sha256.

        Returns:
            an iterable of Content objects matching the search criteria if the
            content exist. Empty iterable otherwise.

        """
        ...

    @remote_api_endpoint("content/get_random")
    def content_get_random(self) -> Sha1Git:
        """Finds a random content id.

        Returns:
            a sha1_git
        """
        ...

    ##########################
    # SkippedContent
    ##########################

    @remote_api_endpoint("content/skipped/add")
    def skipped_content_add(self, content: List[SkippedContent]) -> Dict[str, int]:
        """Add contents to the skipped_content list, which contains
        (partial) information about content missing from the archive.

        Args:
            contents (iterable): iterable of dictionaries representing
                individual pieces of content to add. Each dictionary has the
                following keys:

                - length (Optional[int]): content length (default: -1)
                - one key for each checksum algorithm in
                  :data:`swh.model.hashutil.ALGORITHMS`, mapped to the
                  corresponding checksum; each is optional
                - status (str): must be "absent"
                - reason (str): the reason why the content is absent
                - origin (int): if status = absent, the origin we saw the
                  content in

        Raises:

            The following exceptions can occur:

            - HashCollision in case of collision
            - Any other exceptions raise by the backend

            In case of errors, some content may have been stored in
            the DB and in the objstorage.
            Since additions to both idempotent, that should not be a problem.

        Returns:
            Summary dict with the following key and associated values:

                skipped_content:add: New skipped contents (no data) added
        """
        ...

    @remote_api_endpoint("content/skipped/find")
    def skipped_content_find(self, content: HashDict) -> List[SkippedContent]:
        """Find skipped content for the given hashes

        Args:
            content: a dictionary representing one content hash, mapping
                checksum algorithm names (see swh.model.hashutil.ALGORITHMS) to
                checksum values

        Raises:
            ValueError: in case the key of the dictionary is not sha1, sha1_git
                nor sha256.

        Returns:
            a list of SkippedContent objects matching the search criteria if the
            skipped content exists. Empty list otherwise.
        """
        ...

    @remote_api_endpoint("content/skipped/missing")
    def skipped_content_missing(
        self, contents: List[Dict[str, Any]]
    ) -> Iterable[Dict[str, Any]]:
        """List skipped contents missing from storage.

        Args:
            contents: iterable of dictionaries containing the data for each
                checksum algorithm.

        Returns:
            Iterable of missing skipped contents as dict

        """
        ...

    ##########################
    # Directory
    ##########################

    @remote_api_endpoint("directory/add")
    def directory_add(self, directories: List[Directory]) -> Dict[str, int]:
        """Add directories to the storage

        Args:
            directories (iterable): iterable of dictionaries representing the
                individual directories to add. Each dict has the following
                keys:

                - id (sha1_git): the id of the directory to add
                - entries (list): list of dicts for each entry in the
                      directory.  Each dict has the following keys:

                      - name (bytes)
                      - type (one of 'file', 'dir', 'rev'): type of the
                        directory entry (file, directory, revision)
                      - target (sha1_git): id of the object pointed at by the
                        directory entry
                      - perms (int): entry permissions

        Returns:
            Summary dict of keys with associated count as values:

                directory:add: Number of directories actually added

        """
        ...

    @remote_api_endpoint("directory/missing")
    def directory_missing(self, directories: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List directories missing from storage.

        Args:
            directories: list of directory ids

        Yields:
            missing directory ids

        """
        ...

    @remote_api_endpoint("directory/ls")
    def directory_ls(
        self, directory: Sha1Git, recursive: bool = False
    ) -> Iterable[Dict[str, Any]]:
        """List entries for one directory.

        If `recursive=True`, names in the path of a dir/file not at the
        root are concatenated with a slash (`/`).

        Args:
            directory: the directory to list entries from.
            recursive: if flag on, this list recursively from this directory.

        Yields:
            directory entries for such directory.

        """
        ...

    @remote_api_endpoint("directory/path")
    def directory_entry_get_by_path(
        self, directory: Sha1Git, paths: List[bytes]
    ) -> Optional[Dict[str, Any]]:
        """Get the directory entry (either file or dir) from directory with path.

        Args:
            directory: directory id
            paths: path to lookup from the top level directory. From left
              (top) to right (bottom).

        Returns:
            The corresponding directory entry as dict if found, None otherwise.

        """
        ...

    @remote_api_endpoint("directory/get_entries")
    def directory_get_entries(
        self,
        directory_id: Sha1Git,
        page_token: Optional[bytes] = None,
        limit: int = 1000,
    ) -> Optional[PagedResult[DirectoryEntry]]:
        """Get the content, possibly partial, of a directory with the given id

        The entries of the directory are not guaranteed to be returned in any
        particular order.

        The number of results is not guaranteed to be lower than the ``limit``.

        Args:
            directory_id: identifier of the directory
            page_token: opaque string used to get the next results of a search
            limit: Number of entries to return

        Returns:
            None if the directory does not exist; a page of DirectoryEntry
              objects otherwise.

        See Also:
            :py:func:`swh.storage.algos.directories.directory_get` will get all
            entries for a given directory.
            :py:func:`swh.storage.algos.directories.directory_get_many` will do
            the same for a set of directories.
        """
        ...

    @remote_api_endpoint("directory/get_raw_manifest")
    def directory_get_raw_manifest(
        self, directory_ids: List[Sha1Git]
    ) -> Dict[Sha1Git, Optional[bytes]]:
        """Returns the raw manifest of directories that do not fit the SWH data model,
        or None if they do.
        Directories missing from the archive are not returned at all.

        Args:
            directory_ids: List of directory ids to query
        """
        ...

    @remote_api_endpoint("directory/get_random")
    def directory_get_random(self) -> Sha1Git:
        """Finds a random directory id.

        Returns:
            a sha1_git
        """
        ...

    @remote_api_endpoint("directory/partition/id")
    def directory_get_id_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Sha1Git]:
        """Splits directories into nb_partitions, and returns all the ids and
        raw manifests in one of these based on partition_id (which must be in
        [0, nb_partitions-1]).
        This does not return directory entries themselves; they should be
        retrieved using :meth:`directory_get_entries` and
        :meth:`directory_get_raw_manifest` instead.

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id: index of the partition to fetch
            nb_partitions: total number of partitions to split into

        Returns:
            Page of the directories' sha1_git hashes.
        """
        ...

    ##########################
    # Revision
    ##########################

    @remote_api_endpoint("revision/add")
    def revision_add(self, revisions: List[Revision]) -> Dict[str, int]:
        """Add revisions to the storage

        Args:
            revisions (List[dict]): iterable of dictionaries representing
                the individual revisions to add. Each dict has the following
                keys:

                - **id** (:class:`sha1_git`): id of the revision to add
                - **date** (:class:`dict`): date the revision was written
                - **committer_date** (:class:`dict`): date the revision got
                  added to the origin
                - **type** (one of 'git', 'tar'): type of the
                  revision added
                - **directory** (:class:`sha1_git`): the directory the
                  revision points at
                - **message** (:class:`bytes`): the message associated with
                  the revision
                - **author** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email
                - **committer** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email
                - **metadata** (:class:`jsonb`): extra information as
                  dictionary
                - **synthetic** (:class:`bool`): revision's nature (tarball,
                  directory creates synthetic revision`)
                - **parents** (:class:`list[sha1_git]`): the parents of
                  this revision

        date dictionaries have the form defined in :mod:`swh.model`.

        Returns:
            Summary dict of keys with associated count as values

                revision:add: New objects actually stored in db

        """
        ...

    @remote_api_endpoint("revision/missing")
    def revision_missing(self, revisions: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List revisions missing from storage

        Args:
            revisions: revision ids

        Yields:
            missing revision ids

        """
        ...

    @remote_api_endpoint("revision/partition")
    def revision_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Revision]:
        """Splits revisions into nb_partitions, and returns one of these based on
        partition_id (which must be in [0, nb_partitions-1])

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id: index of the partition to fetch
            nb_partitions: total number of partitions to split into

        Returns:
            Page of Revision model objects within the partition.

        """
        ...

    @remote_api_endpoint("revision")
    def revision_get(
        self, revision_ids: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Revision]]:
        """Get revisions from storage

        Args:
            revisions: revision ids
            ignore_displayname: return the original author/committer's full name even if
              it's masked by a displayname.

        Returns:
            list of revision object (if the revision exists or None otherwise)

        """
        ...

    @remote_api_endpoint("revision/log")
    def revision_log(
        self,
        revisions: List[Sha1Git],
        ignore_displayname: bool = False,
        limit: Optional[int] = None,
    ) -> Iterable[Optional[Dict[str, Any]]]:
        """Fetch revision entry from the given root revisions.

        Args:
            revisions: array of root revisions to lookup
            ignore_displayname: return the original author/committer's full name even if
              it's masked by a displayname.
            limit: limitation on the output result. Default to None.

        Yields:
            revision entries log from the given root root revisions

        """
        ...

    @remote_api_endpoint("revision/shortlog")
    def revision_shortlog(
        self, revisions: List[Sha1Git], limit: Optional[int] = None
    ) -> Iterable[Optional[Tuple[Sha1Git, Tuple[Sha1Git, ...]]]]:
        """Fetch the shortlog for the given revisions

        Args:
            revisions: list of root revisions to lookup
            limit: depth limitation for the output

        Yields:
            a list of (id, parents) tuples

        """
        ...

    @remote_api_endpoint("revision/get_random")
    def revision_get_random(self) -> Sha1Git:
        """Finds a random revision id.

        Returns:
            a sha1_git
        """
        ...

    ##########################
    # ExtID
    ##########################

    @remote_api_endpoint("extid/from_extid")
    def extid_get_from_extid(
        self, id_type: str, ids: List[bytes], version: Optional[int] = None
    ) -> List[ExtID]:
        """Get ExtID objects from external IDs

        Args:
            id_type: type of the given external identifiers (e.g. 'mercurial')
            ids: list of external IDs
            version: (Optional) version to use as filter

        Returns:
            list of ExtID objects

        """
        ...

    @remote_api_endpoint("extid/from_target")
    def extid_get_from_target(
        self,
        target_type: ObjectType,
        ids: List[Sha1Git],
        extid_type: Optional[str] = None,
        extid_version: Optional[int] = None,
    ) -> List[ExtID]:
        """Get ExtID objects from target IDs and target_type

        Args:
            target_type: type the SWH object
            ids: list of target IDs
            extid_type: (Optional) extid_type to use as filter. This cannot be empty if
              extid_version is provided.
            extid_version: (Optional) version to use as filter. This cannot be empty if
              extid_type is provided.

        Raises:
            ValueError if extid_version is provided without extid_type and vice versa.

        Returns:
            list of ExtID objects

        """
        ...

    @remote_api_endpoint("extid/add")
    def extid_add(self, ids: List[ExtID]) -> Dict[str, int]:
        """Add a series of ExtID objects

        Args:
            ids: list of ExtID objects

        Returns:
            Summary dict of keys with associated count as values

                extid:add: New ExtID objects actually stored in db
        """
        ...

    ##########################
    # Release
    ##########################

    @remote_api_endpoint("release/add")
    def release_add(self, releases: List[Release]) -> Dict[str, int]:
        """Add releases to the storage

        Args:
            releases (List[dict]): iterable of dictionaries representing
                the individual releases to add. Each dict has the following
                keys:

                - **id** (:class:`sha1_git`): id of the release to add
                - **revision** (:class:`sha1_git`): id of the revision the
                  release points to
                - **date** (:class:`dict`): the date the release was made
                - **name** (:class:`bytes`): the name of the release
                - **comment** (:class:`bytes`): the comment associated with
                  the release
                - **author** (:class:`Dict[str, bytes]`): dictionary with
                  keys: name, fullname, email

        the date dictionary has the form defined in :mod:`swh.model`.

        Returns:
            Summary dict of keys with associated count as values

                release:add: New objects contents actually stored in db

        """
        ...

    @remote_api_endpoint("release/missing")
    def release_missing(self, releases: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List missing release ids from storage

        Args:
            releases: release ids

        Yields:
            a list of missing release ids

        """
        ...

    @remote_api_endpoint("release")
    def release_get(
        self, releases: List[Sha1Git], ignore_displayname: bool = False
    ) -> List[Optional[Release]]:
        """Given a list of sha1, return the releases's information

        Args:
            releases: list of sha1s
            ignore_displayname: return the original author's full name even if it's
              masked by a displayname.

        Returns:
            List of releases matching the identifiers or None if the release does
            not exist.

        """
        ...

    @remote_api_endpoint("release/get_random")
    def release_get_random(self) -> Sha1Git:
        """Finds a random release id.

        Returns:
            a sha1_git
        """
        ...

    @remote_api_endpoint("release/partition")
    def release_get_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Release]:
        """Splits releases into nb_partitions, and returns one of these based on
        partition_id (which must be in [0, nb_partitions-1])

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id: index of the partition to fetch
            nb_partitions: total number of partitions to split into

        Returns:
            Page of Release model objects within the partition.

        """
        ...

    ##########################
    # Snapshot
    ##########################

    @remote_api_endpoint("snapshot/add")
    def snapshot_add(self, snapshots: List[Snapshot]) -> Dict[str, int]:
        """Add snapshots to the storage.

        Args:
            snapshot ([dict]): the snapshots to add, containing the
              following keys:

              - **id** (:class:`bytes`): id of the snapshot
              - **branches** (:class:`dict`): branches the snapshot contains,
                mapping the branch name (:class:`bytes`) to the branch target,
                itself a :class:`dict` (or ``None`` if the branch points to an
                unknown object)

                - **target_type** (:class:`str`): one of ``content``,
                  ``directory``, ``revision``, ``release``,
                  ``snapshot``, ``alias``
                - **target** (:class:`bytes`): identifier of the target
                  (currently a ``sha1_git`` for all object kinds, or the name
                  of the target branch for aliases)

        Raises:
            ValueError: if the origin or visit id does not exist.

        Returns:

            Summary dict of keys with associated count as values

                snapshot:add: Count of object actually stored in db

        """
        ...

    @remote_api_endpoint("snapshot/missing")
    def snapshot_missing(self, snapshots: List[Sha1Git]) -> Iterable[Sha1Git]:
        """List snapshots missing from storage

        Args:
            snapshots: snapshot ids

        Yields:
            missing snapshot ids

        """
        ...

    @remote_api_endpoint("snapshot")
    def snapshot_get(self, snapshot_id: Sha1Git) -> Optional[Dict[str, Any]]:
        """Get the content, possibly partial, of a snapshot with the given id

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        .. warning:: At most 1000 branches contained in the snapshot will be
            returned for performance reasons. In order to browse the whole
            set of branches, the method :meth:`snapshot_get_branches`
            should be used instead.

        Args:
            snapshot_id: snapshot identifier

        Returns:
            dict: a dict with three keys:
                * **id**: identifier of the snapshot
                * **branches**: a dict of branches contained in the snapshot
                  whose keys are the branches' names.
                * **next_branch**: the name of the first branch not returned
                  or :const:`None` if the snapshot has less than 1000
                  branches.
        """
        ...

    @remote_api_endpoint("snapshot/count_branches")
    def snapshot_count_branches(
        self,
        snapshot_id: Sha1Git,
        branch_name_exclude_prefix: Optional[bytes] = None,
    ) -> Optional[Dict[Optional[str], int]]:
        """Count the number of branches in the snapshot with the given id

        Args:
            snapshot_id: snapshot identifier
            branch_name_exclude_prefix: if provided, do not count branches whose name
                starts with given prefix

        Returns:
            A dict whose keys are the target types of branches and values their
            corresponding amount

        """
        ...

    @remote_api_endpoint("snapshot/get_branches")
    def snapshot_get_branches(
        self,
        snapshot_id: Sha1Git,
        branches_from: bytes = b"",
        branches_count: int = 1000,
        target_types: Optional[List[str]] = None,
        branch_name_include_substring: Optional[bytes] = None,
        branch_name_exclude_prefix: Optional[bytes] = None,
    ) -> Optional[PartialBranches]:
        """Get the content, possibly partial, of a snapshot with the given id

        The branches of the snapshot are iterated in the lexicographical
        order of their names.

        Args:
            snapshot_id: identifier of the snapshot
            branches_from: optional parameter used to skip branches
                whose name is lesser than it before returning them
            branches_count: optional parameter used to restrain
                the amount of returned branches
            target_types: optional parameter used to filter the
                target types of branch to return (possible values that can be
                contained in that list are `'content', 'directory',
                'revision', 'release', 'snapshot', 'alias'`)
            branch_name_include_substring: if provided, only return branches whose name
                contains given substring
            branch_name_exclude_prefix: if provided, do not return branches whose name
                contains given prefix

        Returns:
            a PartialBranches object listing a limited amount of branches
            matching the given criteria or None if the snapshot does not exist.

        See Also:
            :py:func:`swh.storage.algos.snapshot.snapshot_get_all_branches` will get
            all branches for a given snapshot.
        """
        ...

    @remote_api_endpoint("snapshot/get_random")
    def snapshot_get_random(self) -> Sha1Git:
        """Finds a random snapshot id.

        Returns:
            a sha1_git
        """
        ...

    @remote_api_endpoint("snapshot/branches/get_by_name")
    def snapshot_branch_get_by_name(
        self,
        snapshot_id: Sha1Git,
        branch_name: bytes,
        follow_alias_chain: bool = True,
        max_alias_chain_length: int = 100,
    ) -> Optional[SnapshotBranchByNameResponse]:
        """Get a snapshot branch by its name

        Args:
            snapshot_id: Snapshot identifier
            branch_name: Branch name to look for
            follow_alias_chain: If True, find the first non alias branch.
                Return the first branch (alias or non alias) otherwise
            max_alias_chain_length: Maximum number of alias chains to be
                followed before treating the branch as dangling. This has
                no significance when follow_alias_chain is False.

        Returns:
            A SnapshotBranchByNameResponse object
        """
        ...

    @remote_api_endpoint("snapshot/partition/id")
    def snapshot_get_id_partition(
        self,
        partition_id: int,
        nb_partitions: int,
        page_token: Optional[str] = None,
        limit: int = 1000,
    ) -> PagedResult[Sha1Git]:
        """Splits directories into nb_partitions, and returns all the ids and
        raw manifests in one of these based on partition_id (which must be in
        [0, nb_partitions-1]).
        This does not return directory entries themselves; they should be
        retrieved using :meth:`snapshot_get_branches` instead.

        There is no guarantee on how the partitioning is done, or the
        result order.

        Args:
            partition_id: index of the partition to fetch
            nb_partitions: total number of partitions to split into

        Returns:
            Page of the snapshots' sha1_git hashes
        """
        ...

    ##########################
    # OriginVisit and OriginVisitStatus
    ##########################

    @remote_api_endpoint("origin/visit/add")
    def origin_visit_add(self, visits: List[OriginVisit]) -> Iterable[OriginVisit]:
        """Add visits to storage. If the visits have no id, they will be created and assigned
        one. The resulted visits are visits with their visit id set.

        Args:
            visits: List of OriginVisit objects to add

        Raises:
            StorageArgumentException if some origin visit reference unknown origins

        Returns:
            List[OriginVisit] stored

        """
        ...

    @remote_api_endpoint("origin/visit_status/add")
    def origin_visit_status_add(
        self,
        visit_statuses: List[OriginVisitStatus],
    ) -> Dict[str, int]:
        """Add origin visit statuses.

        If there is already a status for the same origin and visit id at the same
        date, the new one will be either dropped or will replace the existing one
        (it is unspecified which one of these two behaviors happens).

        Args:
            visit_statuses: origin visit statuses to add

        Raises: StorageArgumentException if the origin of the visit status is unknown

        """
        ...

    @remote_api_endpoint("origin/visit/get")
    def origin_visit_get(
        self,
        origin: str,
        page_token: Optional[str] = None,
        order: ListOrder = ListOrder.ASC,
        limit: int = 10,
    ) -> PagedResult[OriginVisit]:
        """Retrieve page of OriginVisit information.

        Args:
            origin: The visited origin
            page_token: opaque string used to get the next results of a search
            order: Order on visit id fields to list origin visits (default to asc)
            limit: Number of visits to return

        Raises:
            StorageArgumentException if the order is wrong or the page_token type is
            mistyped.

        Returns: Page of OriginVisit data model objects. if next_page_token is None,
            there is no longer data to retrieve.

        See Also:
            :py:func:`swh.storage.algos.origin.iter_origin_visits` will iterate
            over all OriginVisits for a given origin.
        """
        ...

    @remote_api_endpoint("origin/visit/find_by_date")
    def origin_visit_find_by_date(
        self, origin: str, visit_date: datetime.datetime, type: Optional[str] = None
    ) -> Optional[OriginVisit]:
        """Retrieves the origin visit whose date is closest to the provided
        timestamp.
        In case of a tie, the visit with largest id is selected.

        Args:
            origin: origin (URL)
            visit_date: expected visit date
            type: filter on a specific visit type if provided

        Returns:
            A visit if found, None otherwise

        """
        ...

    @remote_api_endpoint("origin/visit/getby")
    def origin_visit_get_by(self, origin: str, visit: int) -> Optional[OriginVisit]:
        """Retrieve origin visit's information.

        Args:
            origin: origin (URL)
            visit: visit id

        Returns:
            The information on that particular OriginVisit or None if
            it does not exist

        """
        ...

    @remote_api_endpoint("origin/visit/get_latest")
    def origin_visit_get_latest(
        self,
        origin: str,
        type: Optional[str] = None,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
    ) -> Optional[OriginVisit]:
        """Get the latest origin visit for the given origin, optionally
        looking only for those with one of the given allowed_statuses
        or for those with a snapshot.

        Args:
            origin: origin URL
            type: Optional visit type to filter on (e.g git, tar, dsc, svn,
            hg, npm, pypi, ...)
            allowed_statuses: list of visit statuses considered
                to find the latest visit. For instance,
                ``allowed_statuses=['full']`` will only consider visits that
                have successfully run to completion.
            require_snapshot: If True, only a visit with a snapshot
                will be returned.

        Raises:
            StorageArgumentException if values for the allowed_statuses parameters
            are unknown

        Returns:
            OriginVisit matching the criteria if found, None otherwise. Note that as
            OriginVisit no longer held reference on the visit status or snapshot, you
            may want to use origin_visit_status_get_latest for those information.

        """
        ...

    @remote_api_endpoint("origin/visit_status/get")
    def origin_visit_status_get(
        self,
        origin: str,
        visit: int,
        page_token: Optional[str] = None,
        order: ListOrder = ListOrder.ASC,
        limit: int = 10,
    ) -> PagedResult[OriginVisitStatus]:
        """Retrieve page of OriginVisitStatus information.

        Args:
            origin: The visited origin
            visit: The visit identifier
            page_token: opaque string used to get the next results of a search
            order: Order on visit status objects to list (default to asc)
            limit: Number of visit statuses to return

        Returns: Page of OriginVisitStatus data model objects. if next_page_token is
            None, there is no longer data to retrieve.

        See Also:
            :py:func:`swh.storage.algos.origin.iter_origin_visit_statuses` will iterate
            over all OriginVisitStatus objects for a given origin and visit.
        """
        ...

    @remote_api_endpoint("origin/visit_status/get_latest")
    def origin_visit_status_get_latest(
        self,
        origin_url: str,
        visit: int,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
    ) -> Optional[OriginVisitStatus]:
        """Get the latest origin visit status for the given origin visit, optionally
        looking only for those with one of the given allowed_statuses or with a
        snapshot.

        Args:
            origin: origin URL

            allowed_statuses: list of visit statuses considered to find the latest
                visit. Possible values are {created, ongoing, partial, full}. For
                instance, ``allowed_statuses=['full']`` will only consider visits that
                have successfully run to completion.
            require_snapshot: If True, only a visit with a snapshot
                will be returned.

        Raises:
            StorageArgumentException if values for the allowed_statuses parameters
            are unknown

        Returns:
            The OriginVisitStatus matching the criteria

        """
        ...

    @remote_api_endpoint("origin/visit/get_with_statuses")
    def origin_visit_get_with_statuses(
        self,
        origin: str,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
        page_token: Optional[str] = None,
        order: ListOrder = ListOrder.ASC,
        limit: int = 10,
    ) -> PagedResult[OriginVisitWithStatuses]:
        """Retrieve page of origin visits and all their statuses.

        Origin visit statuses are always sorted in ascending order of their dates.

        Args:
            origin: The visited origin URL
            allowed_statuses: Only visit statuses matching that list will be returned.
                If empty, all visit statuses will be returned. Possible status values
                are ``created``, ``not_found``, ``ongoing``, ``failed``, ``partial``
                and ``full``.
            require_snapshot: If :const:`True`, only visit statuses with a snapshot
                will be returned.
            page_token: opaque string used to get the next results
            order: Order on visit objects to list (default to asc)
            limit: Number of visits with their statuses to return

        Returns: Page of OriginVisitWithStatuses objects. if next_page_token is
            None, there is no longer data to retrieve.
        """
        ...

    @remote_api_endpoint("origin/visit_status/get_random")
    def origin_visit_status_get_random(self, type: str) -> Optional[OriginVisitStatus]:
        """Randomly select one successful origin visit with <type>
        made in the last 3 months.

        Returns:
            One random OriginVisitStatus matching the selection criteria

        """
        ...

    ##########################
    # Origin
    ##########################

    @remote_api_endpoint("origin/get")
    def origin_get(self, origins: List[str]) -> List[Optional[Origin]]:
        """Return origins.

        Args:
            origin: a list of urls to find

        Returns:
            the list of associated existing origin model objects. The unknown origins
            will be returned as None at the same index as the input.

        """
        ...

    @remote_api_endpoint("origin/get_sha1")
    def origin_get_by_sha1(self, sha1s: List[bytes]) -> List[Optional[Dict[str, Any]]]:
        """Return origins, identified by the sha1 of their URLs.

        Args:
            sha1s: a list of sha1s

        Returns:
            List of origins dict whose sha1 of their url match, None otherwise.

        """
        ...

    @remote_api_endpoint("origin/list")
    def origin_list(
        self, page_token: Optional[str] = None, limit: int = 100
    ) -> PagedResult[Origin]:
        """Returns the list of origins

        Args:
            page_token: opaque token used for pagination.
            limit: the maximum number of results to return

        Returns:
            Page of Origin data model objects. if next_page_token is None, there is
            no longer data to retrieve.

        """
        ...

    @remote_api_endpoint("origin/search")
    def origin_search(
        self,
        url_pattern: str,
        page_token: Optional[str] = None,
        limit: int = 50,
        regexp: bool = False,
        with_visit: bool = False,
        visit_types: Optional[List[str]] = None,
    ) -> PagedResult[Origin]:
        """Search for origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The search is performed in a case insensitive way.

        Args:
            url_pattern: the string pattern to search for in origin urls
            page_token: opaque token used for pagination
            limit: the maximum number of found origins to return
            regexp: if True, consider the provided pattern as a regular
                expression and return origins whose urls match it
            with_visit: if True, filter out origins with no visit
            visit_types: Only origins having any of the provided visit types
                (e.g. git, svn, pypi) will be returned

        Yields:
            PagedResult of Origin

        """
        ...

    @deprecated
    @remote_api_endpoint("origin/count")
    def origin_count(
        self, url_pattern: str, regexp: bool = False, with_visit: bool = False
    ) -> int:
        """Count origins whose urls contain a provided string pattern
        or match a provided regular expression.
        The pattern search in origin urls is performed in a case insensitive
        way.

        Args:
            url_pattern (str): the string pattern to search for in origin urls
            regexp (bool): if True, consider the provided pattern as a regular
                expression and return origins whose urls match it
            with_visit (bool): if True, filter out origins with no visit

        Returns:
            int: The number of origins matching the search criterion.
        """
        ...

    @remote_api_endpoint("origin/snapshot/get")
    def origin_snapshot_get_all(self, origin_url: str) -> List[Sha1Git]:
        """Return all unique snapshot identifiers resulting from origin visits.

        Args:
            origin_url: origin URL

        Returns:
            list of sha1s

        """
        ...

    @remote_api_endpoint("origin/add_multi")
    def origin_add(self, origins: List[Origin]) -> Dict[str, int]:
        """Add origins to the storage

        Args:
            origins: list of dictionaries representing the individual origins,
                with the following keys:

                - type: the origin type ('git', 'svn', 'deb', ...)
                - url (bytes): the url the origin points to

        Returns:
            Summary dict of keys with associated count as values

                origin:add: Count of object actually stored in db

        """
        ...

    ##########################
    # reverse index for object references
    ##########################

    @remote_api_endpoint("object/find_references")
    def object_find_recent_references(
        self, target_swhid: ExtendedSWHID, limit: int
    ) -> List[ExtendedSWHID]:
        """Return the SWHIDs of objects that are known to reference the object
        ``target_swhid``.

        Args:
            target_swhid: the SWHID of the object targeted by the returned objects
            limit: the maximum number of SWHIDs to return

        Note:
            The data returned by this function is by essence limited to objects that
            were recently added to the archive, and is pruned regularly. For
            completeness, one must also query :mod:`swh.graph` for backwards edges
            targeting the requested object.
        """
        ...

    @remote_api_endpoint("object/references_add")
    def object_references_add(
        self, references: List[ObjectReference]
    ) -> Dict[str, int]:
        """For each object reference ``(source, target)``, record that the ``source``
        object references the ``target`` object (meaning that the ``target`` needs to
        exist for the ``source`` object to be consistent within the archive).

        This function will only be called internally by a reference recording proxy,
        through one of :func:`directory_add`, :func:`revision_add`, :func:`release_add`,
        :func:`snapshot_add`, or :func:`origin_visit_status_add`. External users of
        :mod:`swh.storage` should not need to use this function directly.

        Note:
            these records are inserted in time-based partitions that can be pruned when
            the objects are known in an up-to-date :mod:`swh.graph` instance.

        Args:
            references: a list of ``(source, target)`` SWHID tuples

        Returns:
            A summary dict with the following keys:

              object_reference:add: the number of object references added

        """
        ...

    ##########################
    # misc.
    ##########################

    @remote_api_endpoint("object/find_by_sha1_git")
    def object_find_by_sha1_git(self, ids: List[Sha1Git]) -> Dict[Sha1Git, List[Dict]]:
        """Return the objects found with the given ids.

        Args:
            ids: a generator of sha1_gits

        Returns:
            A dict from id to the list of objects found for that id. Each object
            found is itself a dict with keys:

            - sha1_git: the input id
            - type: the type of object found

        """
        ...

    def stat_counters(self):
        """compute statistics about the number of tuples in various tables

        Returns:
            dict: a dictionary mapping textual labels (e.g., content) to
            integer values (e.g., the number of tuples in table content)

        """
        ...

    def refresh_stat_counters(self):
        """Recomputes the statistics for `stat_counters`."""
        ...

    ##########################
    # RawExtrinsicMetadata
    ##########################

    @remote_api_endpoint("raw_extrinsic_metadata/add")
    def raw_extrinsic_metadata_add(
        self,
        metadata: List[RawExtrinsicMetadata],
    ) -> Dict[str, int]:
        """Add extrinsic metadata on objects (contents, directories, ...).

        The authority and fetcher must be known to the storage before
        using this endpoint.

        If there is already metadata for the same object, authority,
        fetcher, and at the same date; the new one will be either dropped or
        will replace the existing one
        (it is unspecified which one of these two behaviors happens).

        Args:
            metadata: iterable of RawExtrinsicMetadata objects to be inserted.
        """
        ...

    @remote_api_endpoint("raw_extrinsic_metadata/get")
    def raw_extrinsic_metadata_get(
        self,
        target: ExtendedSWHID,
        authority: MetadataAuthority,
        after: Optional[datetime.datetime] = None,
        page_token: Optional[bytes] = None,
        limit: int = 1000,
    ) -> PagedResult[RawExtrinsicMetadata]:
        """Retrieve list of all raw_extrinsic_metadata entries targeting the id

        Args:
            target: the SWHID of the objects to find metadata on
            authority: a dict containing keys `type` and `url`.
            after: minimum discovery_date for a result to be returned
            page_token: opaque token, used to get the next page of results
            limit: maximum number of results to be returned

        Returns:
            PagedResult of RawExtrinsicMetadata

        Raises:
            UnknownMetadataAuthority: if the metadata authority does not exist
                at all
        """
        ...

    @remote_api_endpoint("raw_extrinsic_metadata/get_by_ids")
    def raw_extrinsic_metadata_get_by_ids(
        self, ids: List[Sha1Git]
    ) -> List[RawExtrinsicMetadata]:
        """Retrieve list of raw_extrinsic_metadata entries of the given id
        (unlike raw_extrinsic_metadata_get, which returns metadata entries
        **targeting** the id)

        Args:
            ids: list of hashes of RawExtrinsicMetadata objects

        """
        ...

    @remote_api_endpoint("raw_extrinsic_metadata/get_authorities")
    def raw_extrinsic_metadata_get_authorities(
        self, target: ExtendedSWHID
    ) -> List[MetadataAuthority]:
        """Returns all authorities that provided metadata on the given object."""
        ...

    ##########################
    # MetadataFetcher and MetadataAuthority
    ##########################

    @remote_api_endpoint("metadata_fetcher/add")
    def metadata_fetcher_add(
        self,
        fetchers: List[MetadataFetcher],
    ) -> Dict[str, int]:
        """Add new metadata fetchers to the storage.

        Their `name` and `version` together are unique identifiers of this
        fetcher; and `metadata` is an arbitrary dict of JSONable data
        with information about this fetcher, which must not be `None`
        (but may be empty).

        Args:
            fetchers: iterable of MetadataFetcher to be inserted

        """
        ...

    @remote_api_endpoint("metadata_fetcher/get")
    def metadata_fetcher_get(
        self, name: str, version: str
    ) -> Optional[MetadataFetcher]:
        """Retrieve information about a fetcher

        Args:
            name: the name of the fetcher
            version: version of the fetcher

        Returns:
            a MetadataFetcher object (with a non-None metadata field) if it is known,
            else None.

        """
        ...

    @remote_api_endpoint("metadata_authority/add")
    def metadata_authority_add(
        self, authorities: List[MetadataAuthority]
    ) -> Dict[str, int]:
        """Add new metadata authorities to the storage.

        Their `type` and `url` together are unique identifiers of this
        authority; and `metadata` is an arbitrary dict of JSONable data
        with information about this authority, which must not be `None`
        (but may be empty).

        Args:
            authorities: iterable of MetadataAuthority to be inserted
        """
        ...

    @remote_api_endpoint("metadata_authority/get")
    def metadata_authority_get(
        self, type: MetadataAuthorityType, url: str
    ) -> Optional[MetadataAuthority]:
        """Retrieve information about an authority

        Args:
            type: one of "deposit_client", "forge", or "registry"
            url: unique URI identifying the authority

        Returns:
            a MetadataAuthority object (with a non-None metadata field) if it is known,
            else None.
        """
        ...

    @remote_api_endpoint("clear/buffer")
    def clear_buffers(self, object_types: Sequence[str] = ()) -> None:
        """For backend storages (pg, storage, in-memory), this is a noop operation. For proxy
        storages (especially filter, buffer), this is an operation which cleans internal
        state.

        """

    @remote_api_endpoint("flush")
    def flush(self, object_types: Sequence[str] = ()) -> Dict[str, int]:
        """For backend storages (pg, storage, in-memory), this is expected to be a noop
        operation. For proxy storages (especially buffer), this is expected to trigger
        actual writes to the backend.
        """
        ...


class ObjectDeletionInterface(Protocol):
    def object_delete(self, swhids: List[ExtendedSWHID]) -> Dict[str, int]:
        """Delete objects from the storage

        All skipped content objects matching the given SWHID will be removed,
        including those who have the same SWHID due to hash collisions.

        Origin objects are removed alongside their associated origin visit and
        origin visit status objects.

        Only objects from this facility will be removed. The same method
        should be called on other storage, objstorage, or journal instances
        where the specified objects need to be removed.

        Args:
            swhids: list of SWHID of the objects to remove

        Returns:
            dict: number of objects removed. Details of each key:

            content:delete
                Number of content objects removed

            content:delete:bytes
                Sum of the removed contents’ data length

            skipped_content:delete
                Number of skipped content objects removed

            directory:delete
                Number of directory objects removed

            revision:delete
                Number of revision objects removed

            release:delete
                Number of release objects removed

            snapshot:delete
                Number of snapshot objects removed

            origin:delete
                Number of origin objects removed

            origin_visit:delete
                Number of origin visit objects removed

            origin_visit_status:delete
                Number of origin visit status objects removed

            ori_metadata:delete
                Number of raw extrinsic metadata objects targeting
                an origin that have been removed

            snp_metadata:delete
                Number of raw extrinsic metadata objects targeting
                a snapshot that have been removed

            rev_metadata:delete
                Number of raw extrinsic metadata objects targeting
                a revision that have been removed

            rel_metadata:delete
                Number of raw extrinsic metadata objects targeting
                a release that have been removed

            dir_metadata:delete
                Number ef raw extrinsic metadata objects targeting
                a directory that have been removed

            cnt_metadata:delete
                Number of raw extrinsic metadata objects targeting
                a content that have been removed

            emd_metadata:delete
                Number of raw extrinsic metadata objects targeting
                a raw extrinsic metadata object that have been removed"""
        ...

    def extid_delete_for_target(self, target_swhids: List[CoreSWHID]) -> Dict[str, int]:
        """Delete ExtID objects from the storage

        Args:
            target_swhids: list of SWHIDs targeted by the ExtID objects to remove

        Returns:
            Summary dict with the following keys and associated values:

                extid:delete: Number of ExtID objects removed
        """
        ...



# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from functools import partial
import os
import resource
import signal
import socket
import subprocess
import time

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.storage import get_storage
from swh.storage.postgresql.db import Db as PostgreSQLDb
from swh.storage.postgresql.storage import Storage as StorageDatastore
from swh.storage.tests.storage_data import StorageData

os.environ["LC_ALL"] = "C.UTF-8"


_CASSANDRA_CONFIG_TEMPLATE = """
data_file_directories:
    - {data_dir}/data
commitlog_directory: {data_dir}/commitlog
hints_directory: {data_dir}/hints
saved_caches_directory: {data_dir}/saved_caches

commitlog_sync: periodic
commitlog_sync_period_in_ms: 1000000
partitioner: org.apache.cassandra.dht.Murmur3Partitioner
endpoint_snitch: SimpleSnitch
seed_provider:
    - class_name: org.apache.cassandra.locator.SimpleSeedProvider
      parameters:
          - seeds: "127.0.0.1"

storage_port: {storage_port}
native_transport_port: {native_transport_port}
start_native_transport: true
listen_address: 127.0.0.1

enable_user_defined_functions: true

# speed-up by disabling period saving to disk
key_cache_save_period: 0
row_cache_save_period: 0
trickle_fsync: false
commitlog_sync_period_in_ms: 100000
authenticator: PasswordAuthenticator
"""

_SCYLLA_EXTRA_CONFIG_TEMPLATE = """
experimental_features:
    - udf
view_hints_directory: {data_dir}/view_hints
prometheus_port: 0  # disable prometheus server
start_rpc: false  # disable thrift server
api_port: {api_port}
"""


def _free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _wait_for_peer(addr, port):
    wait_until = time.time() + 60
    while time.time() < wait_until:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((addr, port))
        except ConnectionRefusedError:
            time.sleep(0.1)
        else:
            sock.close()
            return True
    return False


@pytest.fixture(scope="session")
def cassandra_auth_provider_config():
    return {
        "cls": "cassandra.auth.PlainTextAuthProvider",
        "username": "cassandra",
        "password": "cassandra",
    }


@pytest.fixture(scope="session")
def swh_storage_cassandra_cluster(tmpdir_factory):
    cassandra_conf = tmpdir_factory.mktemp("cassandra_conf")
    cassandra_data = tmpdir_factory.mktemp("cassandra_data")
    cassandra_log = tmpdir_factory.mktemp("cassandra_log")
    native_transport_port = _free_port()
    storage_port = _free_port()
    jmx_port = _free_port()
    api_port = _free_port()

    use_scylla = bool(os.environ.get("SWH_USE_SCYLLADB", ""))

    cassandra_bin = os.environ.get(
        "SWH_CASSANDRA_BIN", "/usr/bin/scylla" if use_scylla else "/usr/sbin/cassandra"
    )

    if use_scylla:
        os.makedirs(cassandra_conf.join("conf"))
        config_path = cassandra_conf.join("conf/scylla.yaml")
        config_template = _CASSANDRA_CONFIG_TEMPLATE + _SCYLLA_EXTRA_CONFIG_TEMPLATE
    else:
        config_path = cassandra_conf.join("cassandra.yaml")
        config_template = _CASSANDRA_CONFIG_TEMPLATE

    with open(str(config_path), "w") as fd:
        fd.write(
            config_template.format(
                data_dir=str(cassandra_data),
                storage_port=storage_port,
                native_transport_port=native_transport_port,
                api_port=api_port,
            )
        )

    if os.environ.get("SWH_CASSANDRA_LOG"):
        stdout = stderr = None
    else:
        stdout = stderr = subprocess.DEVNULL

    env = {
        "MAX_HEAP_SIZE": "300M",
        "HEAP_NEWSIZE": "50M",
        "JVM_OPTS": "-Xlog:gc=error:file=%s/gc.log" % cassandra_log,
    }
    if "JAVA_HOME" in os.environ:
        env["JAVA_HOME"] = os.environ["JAVA_HOME"]

    if use_scylla:
        env = {
            **env,
            "SCYLLA_HOME": cassandra_conf,
        }
        # prevent "NOFILE rlimit too low (recommended setting 200000,
        # minimum setting 10000; refusing to start."
        resource.setrlimit(resource.RLIMIT_NOFILE, (200000, 200000))

        proc = subprocess.Popen(
            [
                cassandra_bin,
                "--developer-mode=1",
            ],
            start_new_session=True,
            env=env,
            stdout=stdout,
            stderr=stderr,
        )
    else:
        proc = subprocess.Popen(
            [
                cassandra_bin,
                "-Dcassandra.config=file://%s/cassandra.yaml" % cassandra_conf,
                "-Dcassandra.logdir=%s" % cassandra_log,
                "-Dcassandra.jmx.local.port=%d" % jmx_port,
                "-Dcassandra-foreground=yes",
            ],
            start_new_session=True,
            env=env,
            stdout=stdout,
            stderr=stderr,
        )

    listening = _wait_for_peer("127.0.0.1", native_transport_port)

    if listening:
        # Wait for initialization
        auth_provider = PlainTextAuthProvider(
            username="cassandra", password="cassandra"
        )
        cluster = Cluster(
            ["127.0.0.1"],
            port=native_transport_port,
            auth_provider=auth_provider,
            connect_timeout=30,
            control_connection_timeout=30,
        )

        session = None
        retry = 0
        while (not session) and retry < 10:
            try:
                session = cluster.connect()
            except Exception:
                time.sleep(1)
                retry += 1
        cluster.shutdown()

        yield (["127.0.0.1"], native_transport_port)

    if not listening or os.environ.get("SWH_CASSANDRA_LOG"):
        debug_log_path = str(cassandra_log.join("debug.log"))
        if os.path.exists(debug_log_path):
            with open(debug_log_path) as fd:
                print(fd.read())

    if not listening:
        if proc.poll() is None:
            raise Exception("cassandra process unexpectedly not listening.")
        else:
            raise Exception("cassandra process unexpectedly stopped.")

    pgrp = os.getpgid(proc.pid)
    os.killpg(pgrp, signal.SIGKILL)


@pytest.fixture(scope="session")
def swh_storage_cassandra_keyspace(
    swh_storage_cassandra_cluster, cassandra_auth_provider_config
):
    from swh.storage.cassandra import create_keyspace

    (hosts, port) = swh_storage_cassandra_cluster
    keyspace = "test" + os.urandom(10).hex()

    create_keyspace(hosts, keyspace, port, auth_provider=cassandra_auth_provider_config)

    return keyspace


@pytest.fixture
def swh_storage_cassandra_backend_config(
    swh_storage_cassandra_cluster,
    swh_storage_cassandra_keyspace,
    cassandra_auth_provider_config,
):
    from swh.storage.cassandra.schema import TABLES

    (hosts, port) = swh_storage_cassandra_cluster

    keyspace = swh_storage_cassandra_keyspace

    storage_config = dict(
        cls="cassandra",
        hosts=hosts,
        port=port,
        keyspace=keyspace,
        journal_writer={"cls": "memory"},
        objstorage={"cls": "memory"},
        auth_provider=cassandra_auth_provider_config,
    )

    yield storage_config

    storage = get_storage(**storage_config)

    for table in TABLES:
        table_rows = storage._cql_runner._session.execute(
            f"SELECT * from {keyspace}.{table} LIMIT 1"
        )
        if table_rows.one() is not None:
            storage._cql_runner._session.execute(f"TRUNCATE TABLE {keyspace}.{table}")

    storage._cql_runner._cluster.shutdown()


def create_object_references_partition(**kwargs):
    db = PostgreSQLDb.connect(**kwargs)
    with db.transaction() as cur:
        db.object_references_create_partition(
            *datetime.date.today().isocalendar()[0:2], cur=cur
        )


swh_storage_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(
            initialize_database_for_module,
            modname="storage",
            version=StorageDatastore.current_version,
        ),
        create_object_references_partition,
    ],
)


swh_storage_postgresql = factories.postgresql(
    "swh_storage_postgresql_proc",
)


@pytest.fixture
def swh_storage_postgresql_backend_config(swh_storage_postgresql):
    """Basic pg storage configuration with no journal collaborator
    (to avoid pulling optional dependency on clients of this fixture)

    """
    yield {
        "cls": "postgresql",
        "db": swh_storage_postgresql.info.dsn,
        "objstorage": {"cls": "memory"},
        "check_config": {"check_write": True},
        "max_pool_conns": 100,
    }


@pytest.fixture
def swh_storage_backend_config(swh_storage_postgresql_backend_config):
    """Configuration to use for :func:`swh_storage_backend`.

    Defaults to :func:`swh_storage_postgresql_backend_config`.

    """
    return swh_storage_postgresql_backend_config


@pytest.fixture
def swh_storage_backend(swh_storage_backend_config):
    """
    By default, this fixture aliases ``swh_storage``. However, when ``swh_storage``
    is overridden to be a proxy storage, this fixture returns the storage instance
    behind all proxies.

    This is useful to introspect the state of backends from proxy tests"""
    return get_storage(**swh_storage_backend_config)


@pytest.fixture
def swh_storage(swh_storage_backend):
    return swh_storage_backend


@pytest.fixture
def sample_data() -> StorageData:
    """Pre-defined sample storage object data to manipulate

    Returns:
        StorageData whose attribute keys are data model objects. Either multiple
        objects: contents, directories, revisions, releases, ... or simple ones:
        content, directory, revision, release, ...

    """
    return StorageData()



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import pytest

from swh.model.model import (
    ObjectType,
    Person,
    Release,
    Revision,
    RevisionType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.swhids import ExtendedSWHID
from swh.storage.postgresql import converters


@pytest.mark.parametrize(
    "model_date,db_date",
    [
        (
            None,
            {
                "timestamp": None,
                "offset": 0,
                "neg_utc_offset": None,
                "offset_bytes": None,
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=1234567890,
                    microseconds=0,
                ),
                offset_bytes=b"+0200",
            ),
            {
                "timestamp": "2009-02-13T23:31:30+00:00",
                "offset": 120,
                "neg_utc_offset": False,
                "offset_bytes": b"+0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=1123456789,
                    microseconds=0,
                ),
                offset_bytes=b"-0000",
            ),
            {
                "timestamp": "2005-08-07T23:19:49+00:00",
                "offset": 0,
                "neg_utc_offset": True,
                "offset_bytes": b"-0000",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=1234567890,
                    microseconds=0,
                ),
                offset_bytes=b"+0042",
            ),
            {
                "timestamp": "2009-02-13T23:31:30+00:00",
                "offset": 42,
                "neg_utc_offset": False,
                "offset_bytes": b"+0042",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=1634366813,
                    microseconds=0,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "2021-10-16T06:46:53+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=0,
                    microseconds=0,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1970-01-01T00:00:00+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=0,
                    microseconds=1,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1970-01-01T00:00:00.000001+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=-1,
                    microseconds=0,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1969-12-31T23:59:59+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=-1,
                    microseconds=1,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1969-12-31T23:59:59.000001+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=-3600,
                    microseconds=0,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1969-12-31T23:00:00+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=-3600,
                    microseconds=1,
                ),
                offset_bytes=b"-0200",
            ),
            {
                "timestamp": "1969-12-31T23:00:00.000001+00:00",
                "offset": -120,
                "neg_utc_offset": False,
                "offset_bytes": b"-0200",
            },
        ),
        (
            TimestampWithTimezone(
                timestamp=Timestamp(
                    seconds=1234567890,
                    microseconds=0,
                ),
                offset_bytes=b"+200",
            ),
            {
                "timestamp": "2009-02-13T23:31:30+00:00",
                "offset": 120,
                "neg_utc_offset": False,
                "offset_bytes": b"+200",
            },
        ),
    ],
)
def test_date(model_date, db_date):
    assert converters.date_to_db(model_date) == db_date
    assert (
        converters.db_to_date(
            date=(
                None
                if db_date["timestamp"] is None
                else datetime.datetime.fromisoformat(db_date["timestamp"])
            ),
            offset_bytes=db_date["offset_bytes"],
        )
        == model_date
    )


def test_db_to_author():
    # when
    actual_author = converters.db_to_author(b"name <email> ", b"name", b"email")

    # then
    assert actual_author == Person.from_fullname(b"name <email> ")


def test_db_to_author_none():
    # when
    actual_author = converters.db_to_author(None, None, None)

    # then
    assert actual_author is None


def test_db_to_author_unparsed():
    author = converters.db_to_author(b"Fullname <email@example.com>", None, None)
    assert author == Person.from_fullname(b"Fullname <email@example.com>")


@pytest.mark.parametrize("tested_func", ["db_to_revision", "db_to_optional_revision"])
def test_db_to_revision(tested_func):
    # when
    actual_revision = getattr(converters, tested_func)(
        {
            "id": b"revision-id",
            "date": None,
            "date_offset": None,
            "date_neg_utc_offset": None,
            "date_offset_bytes": None,
            "committer_date": None,
            "committer_date_offset": None,
            "committer_date_neg_utc_offset": None,
            "committer_date_offset_bytes": None,
            "type": "git",
            "directory": b"dir-sha1",
            "message": b"commit message",
            "author_fullname": b"auth-name <auth-email>",
            "author_name": b"auth-name",
            "author_email": b"auth-email",
            "committer_fullname": b"comm-name <comm-email>",
            "committer_name": b"comm-name",
            "committer_email": b"comm-email",
            "metadata": {},
            "synthetic": False,
            "extra_headers": (),
            "raw_manifest": None,
            "parents": [b"123", b"456"],
        }
    )

    # then
    assert actual_revision == Revision(
        id=b"revision-id",
        author=Person(
            fullname=b"auth-name <auth-email>",
            name=b"auth-name",
            email=b"auth-email",
        ),
        date=None,
        committer=Person(
            fullname=b"comm-name <comm-email>",
            name=b"comm-name",
            email=b"comm-email",
        ),
        committer_date=None,
        type=RevisionType.GIT,
        directory=b"dir-sha1",
        message=b"commit message",
        metadata={},
        synthetic=False,
        extra_headers=(),
        parents=(b"123", b"456"),
    )


@pytest.mark.parametrize("tested_func", ["db_to_revision", "db_to_optional_revision"])
def test_db_to_revision_none(tested_func):
    # when
    row = {
        "id": b"revision-id",
        "date": None,
        "date_offset": None,
        "date_neg_utc_offset": None,
        "date_offset_bytes": None,
        "committer_date": None,
        "committer_date_offset": None,
        "committer_date_neg_utc_offset": None,
        "committer_date_offset_bytes": None,
        "type": None,
        "directory": None,
        "message": None,
        "author_fullname": None,
        "author_name": None,
        "author_email": None,
        "committer_fullname": None,
        "committer_name": None,
        "committer_email": None,
        "metadata": None,
        "synthetic": None,
        "extra_headers": None,
        "raw_manifest": None,
        "parents": [],
    }

    if tested_func == "db_to_revision":
        with pytest.raises(ValueError):
            converters.db_to_revision(row)
    else:
        assert converters.db_to_optional_revision(row) is None


@pytest.mark.parametrize("tested_func", ["db_to_release", "db_to_optional_release"])
def test_db_to_release(tested_func):
    # when
    actual_release = getattr(converters, tested_func)(
        {
            "id": b"release-id",
            "target": b"revision-id",
            "target_type": "revision",
            "date": None,
            "date_offset": None,
            "date_neg_utc_offset": None,
            "date_offset_bytes": None,
            "name": b"release-name",
            "comment": b"release comment",
            "synthetic": True,
            "author_fullname": b"auth-name <auth-email>",
            "author_name": b"auth-name",
            "author_email": b"auth-email",
            "raw_manifest": None,
        }
    )

    # then
    assert actual_release == Release(
        author=Person(
            fullname=b"auth-name <auth-email>",
            name=b"auth-name",
            email=b"auth-email",
        ),
        date=None,
        id=b"release-id",
        name=b"release-name",
        message=b"release comment",
        synthetic=True,
        target=b"revision-id",
        target_type=ObjectType.REVISION,
    )


@pytest.mark.parametrize("tested_func", ["db_to_release", "db_to_optional_release"])
def test_db_to_release_none(tested_func):
    row = {
        "id": b"release-id",
        "target": None,
        "target_type": None,
        "date": None,
        "date_offset": None,
        "date_neg_utc_offset": None,
        "date_offset_bytes": None,
        "name": None,
        "comment": None,
        "synthetic": None,
        "author_fullname": None,
        "author_name": None,
        "author_email": None,
        "raw_manifest": None,
    }

    if tested_func == "db_to_release":
        with pytest.raises(ValueError):
            converters.db_to_release(row)
    else:
        assert converters.db_to_optional_release(row) is None


def test_db_to_raw_extrinsic_metadata_raw_target():
    row = {
        "raw_extrinsic_metadata.target": "https://example.com/origin",
        "metadata_authority.type": "forge",
        "metadata_authority.url": "https://example.com",
        "metadata_fetcher.name": "swh.lister",
        "metadata_fetcher.version": "1.0.0",
        "discovery_date": datetime.datetime(
            2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
        ),
        "format": "text/plain",
        "raw_extrinsic_metadata.metadata": b"metadata",
        "origin": None,
        "visit": None,
        "snapshot": None,
        "release": None,
        "revision": None,
        "path": None,
        "directory": None,
    }

    with pytest.deprecated_call():
        computed_rem = converters.db_to_raw_extrinsic_metadata(row)

    assert computed_rem.target == ExtendedSWHID.from_string(
        "swh:1:ori:5a7439b0b93a5d230b6a67b8e7e0f7dc3c9f6c70"
    )



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import TYPE_CHECKING, Dict, List

from swh.model.hashutil import hash_to_hex
from swh.storage.utils import content_bytes_hashes, content_hex_hashes

if TYPE_CHECKING:
    from swh.model.swhids import ExtendedSWHID
    from swh.storage.proxies.blocking.db import BlockingStatus
    from swh.storage.proxies.masking.db import MaskedStatus


class StorageDBError(Exception):
    """Specific storage db error (connection, erroneous queries, etc...)"""

    def __str__(self):
        return "An unexpected error occurred in the backend: %s" % (self.args,)


class StorageAPIError(Exception):
    """Specific internal storage api (mainly connection)"""

    def __str__(self):
        args = self.args
        return "An unexpected error occurred in the api backend: %s" % (args,)


class NonRetryableException(Exception):
    """Persistent storage exceptions for which clients should not issue automatic
    retries"""

    pass


class StorageArgumentException(NonRetryableException):
    """Argument passed to a Storage endpoint is invalid."""

    pass


class QueryTimeout(StorageAPIError):
    """Raised when a query to the backend (Cassandra or PostgreSQL) took too long."""

    pass


class UnknownMetadataAuthority(StorageArgumentException):
    """Raised when ``raw_extrinsic_metadata_add`` is called with a non-existing
    metadata authority as argument."""

    pass


class UnknownMetadataFetcher(StorageArgumentException):
    """Raised when ``raw_extrinsic_metadata_add`` is called with a non-existing
    metadata fetcher as argument."""

    pass


class BlockedOriginException(NonRetryableException):
    """Raised when the blocking proxy prevent from inserting a blocked origin"""

    def __init__(self, blocked: "Dict[str, BlockingStatus]"):
        blocked = {
            url: status
            for url, status in blocked.items()
            if status.state.name != "NON_BLOCKED"
        }
        if not blocked:
            raise ValueError(
                "Can't raise a BlockedOriginException if no origin is actually blocked"
            )

        self.blocked = blocked
        super().__init__(blocked)

    def __str__(self):
        return "Some origins are blocked: %s" % ", ".join(self.blocked)


class MaskedObjectException(NonRetryableException):
    """Raised when the masking proxy attempts to return a masked object"""

    def __init__(self, masked: "Dict[ExtendedSWHID, List[MaskedStatus]]"):
        if not masked:
            raise ValueError(
                "Can't raise a MaskedObjectException if no objects are masked"
            )
        for swhid, statuses in masked.items():
            if not statuses:
                raise ValueError(f"MaskedObjectException has no statuses for {swhid}")

        self.masked = masked

        super().__init__(masked)

    @staticmethod
    def _str_one_masked(obj, statuses):
        status = statuses[0]
        return f"{obj} by request {status.request} ({status.state.name})" + (
            " and others" if len(statuses) > 1 else ""
        )

    def __str__(self):
        return "Some objects are masked: %s" % ", ".join(
            self._str_one_masked(obj, statuses)
            for obj, statuses in self.masked.items()
            if statuses
        )


class HashCollision(Exception):
    """Exception raised when a content collides in a storage backend"""

    def __init__(self, algo, hash_id, colliding_contents):
        self.algo = algo
        self.hash_id = hash_to_hex(hash_id)
        self.colliding_contents = [content_hex_hashes(c) for c in colliding_contents]
        super().__init__(self.algo, self.hash_id, self.colliding_contents)

    def colliding_content_hashes(self) -> List[Dict[str, bytes]]:
        return [content_bytes_hashes(c) for c in self.colliding_contents]



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from typing import Tuple

import attr

from swh.model import from_disk
from swh.model.hashutil import hash_to_bytes
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ExtID,
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    ObjectType,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Person,
    RawExtrinsicMetadata,
    Release,
    Revision,
    RevisionType,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    Timestamp,
    TimestampWithTimezone,
)
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.model.swhids import ObjectType as SwhidObjectType


class StorageData:
    """Data model objects to use within tests."""

    content = Content(
        data=b"42\n",
        length=3,
        sha1=hash_to_bytes("34973274ccef6ab4dfaaf86599792fa9c3fe4689"),
        sha1_git=hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
        sha256=hash_to_bytes(
            "084c799cd551dd1d8d5c5f9a5d593b2e931f5e36122ee5c793c1d08a19839cc0"
        ),
        blake2s256=hash_to_bytes(
            "d5fe1939576527e42cfd76a9455a2432fe7f56669564577dd93c4280e76d661d"
        ),
        status="visible",
    )
    content2 = Content(
        data=b"4242\n",
        length=5,
        sha1=hash_to_bytes("61c2b3a30496d329e21af70dd2d7e097046d07b7"),
        sha1_git=hash_to_bytes("36fade77193cb6d2bd826161a0979d64c28ab4fa"),
        sha256=hash_to_bytes(
            "859f0b154fdb2d630f45e1ecae4a862915435e663248bb8461d914696fc047cd"
        ),
        blake2s256=hash_to_bytes(
            "849c20fad132b7c2d62c15de310adfe87be94a379941bed295e8141c6219810d"
        ),
        status="visible",
    )
    content3 = Content(
        data=b"424242\n",
        length=7,
        sha1=hash_to_bytes("3e21cc4942a4234c9e5edd8a9cacd1670fe59f13"),
        sha1_git=hash_to_bytes("c932c7649c6dfa4b82327d121215116909eb3bea"),
        sha256=hash_to_bytes(
            "92fb72daf8c6818288a35137b72155f507e5de8d892712ab96277aaed8cf8a36"
        ),
        blake2s256=hash_to_bytes(
            "76d0346f44e5a27f6bafdd9c2befd304aff83780f93121d801ab6a1d4769db11"
        ),
        status="visible",
        ctime=datetime.datetime(2019, 12, 1, tzinfo=datetime.timezone.utc),
    )
    contents: Tuple[Content, ...] = (content, content2, content3)

    skipped_content = SkippedContent(
        length=1024 * 1024 * 200,
        sha1_git=hash_to_bytes("33e45d56f88993aae6a0198013efa80716fd8920"),
        sha1=hash_to_bytes("43e45d56f88993aae6a0198013efa80716fd8920"),
        sha256=hash_to_bytes(
            "7bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
        ),
        blake2s256=hash_to_bytes(
            "ade18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b"
        ),
        reason="Content too long",
        status="absent",
        origin="file:///dev/zero",
    )
    skipped_content2 = SkippedContent(
        length=1024 * 1024 * 300,
        sha1_git=hash_to_bytes("44e45d56f88993aae6a0198013efa80716fd8921"),
        sha1=hash_to_bytes("54e45d56f88993aae6a0198013efa80716fd8920"),
        sha256=hash_to_bytes(
            "8cbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
        ),
        blake2s256=hash_to_bytes(
            "9ce18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b"
        ),
        reason="Content too long",
        status="absent",
    )
    skipped_contents: Tuple[SkippedContent, ...] = (skipped_content, skipped_content2)

    directory5 = Directory(
        id=hash_to_bytes("4b825dc642cb6eb9a060e54bf8d69288fbee4904"),
        entries=(),
    )
    directory = Directory(
        id=hash_to_bytes("5256e856a0a0898966d6ba14feb4388b8b82d302"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"foo",
                    type="file",
                    target=content.sha1_git,
                    perms=from_disk.DentryPerms.content,
                ),
                DirectoryEntry(
                    name=b"bar\xc3",
                    type="dir",
                    target=directory5.id,
                    perms=from_disk.DentryPerms.directory,
                ),
            ],
        ),
    )
    directory2 = Directory(
        id=hash_to_bytes("8505808532953da7d2581741f01b29c04b1cb9ab"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"oof",
                    type="file",
                    target=content2.sha1_git,
                    perms=from_disk.DentryPerms.content,
                )
            ],
        ),
    )
    directory3 = Directory(
        id=hash_to_bytes("13089e6e544f78df7c9a40a3059050d10dee686a"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"foo",
                    type="file",
                    target=content.sha1_git,
                    perms=from_disk.DentryPerms.content,
                ),
                DirectoryEntry(
                    name=b"subdir",
                    type="dir",
                    target=directory.id,
                    perms=from_disk.DentryPerms.directory,
                ),
                DirectoryEntry(
                    name=b"hello",
                    type="file",
                    target=content2.sha1_git,
                    perms=from_disk.DentryPerms.content,
                ),
            ],
        ),
    )
    directory4 = Directory(
        id=hash_to_bytes("cd5dfd9c09d9e99ed123bc7937a0d5fddc3cd531"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"subdir1",
                    type="dir",
                    target=directory3.id,
                    perms=from_disk.DentryPerms.directory,
                )
            ],
        ),
    )

    directory6 = Directory(
        id=hash_to_bytes("afa0105cfcaa14fdbacee344e96659170bb1bda5"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"foo",
                    type="file",
                    target=b"\x00" * 20,
                    perms=from_disk.DentryPerms.content,
                ),
                DirectoryEntry(
                    name=b"bar",
                    type="dir",
                    target=b"\x01" * 20,
                    perms=from_disk.DentryPerms.directory,
                ),
            ],
        ),
        raw_manifest=(
            b"tree 61\x00"
            b"100644 foo\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"  # noqa
            b"40000 bar\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01"  # noqa
        ),
    )

    directory7 = Directory(
        id=hash_to_bytes("a720da7e7ba5af69498f884f91ca7f4ed16997be"),
        entries=tuple(
            [
                DirectoryEntry(
                    name=b"subdir1",
                    type="dir",
                    target=directory5.id,
                    perms=from_disk.DentryPerms.directory,
                ),
                DirectoryEntry(
                    name=b"subdir2",
                    type="dir",
                    target=directory5.id,
                    perms=from_disk.DentryPerms.directory,
                ),
                DirectoryEntry(
                    name=b"subdir3",
                    type="dir",
                    target=directory4.id,
                    perms=from_disk.DentryPerms.directory,
                ),
            ],
        ),
    )

    directories: Tuple[Directory, ...] = (
        directory2,
        directory,
        directory3,
        directory4,
        directory5,
        directory6,
        directory7,
    )

    revision = Revision(
        id=hash_to_bytes("01a7114f36fddd5ef2511b2cadda237a68adbb12"),
        message=b"hello",
        author=Person(
            name=b"Nicolas Dandrimont",
            email=b"nicolas@example.com",
            fullname=b"Nicolas Dandrimont <nicolas@example.com> ",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1234567890, microseconds=0),
            offset_bytes=b"+0200",
        ),
        committer=Person(
            name=b"St\xc3fano Zacchiroli",
            email=b"stefano@example.com",
            fullname=b"St\xc3fano Zacchiroli <stefano@example.com>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1123456789, microseconds=0),
            offset_bytes=b"+0200",
        ),
        parents=(),
        type=RevisionType.GIT,
        directory=directory.id,
        metadata={
            "checksums": {
                "sha1": "tarball-sha1",
                "sha256": "tarball-sha256",
            },
            "signed-off-by": "some-dude",
        },
        extra_headers=(
            (b"gpgsig", b"test123"),
            (b"mergetag", b"foo\\bar"),
            (b"mergetag", b"\x22\xaf\x89\x80\x01\x00"),
        ),
        synthetic=True,
    )
    revision2 = Revision(
        id=hash_to_bytes("a646dd94c912829659b22a1e7e143d2fa5ebde1b"),
        message=b"hello again",
        author=Person(
            name=b"Roberto Dicosmo",
            email=b"roberto@example.com",
            fullname=b"Roberto Dicosmo <roberto@example.com>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"tony",
            email=b"ar@dumont.fr",
            fullname=b"tony <ar@dumont.fr>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1123456789,
                microseconds=220000,
            ),
            offset_bytes=b"+0000",
        ),
        parents=tuple([revision.id]),
        type=RevisionType.GIT,
        directory=directory2.id,
        metadata=None,
        extra_headers=(),
        synthetic=False,
    )
    revision3 = Revision(
        id=hash_to_bytes("beb2844dff30658e27573cb46eb55980e974b391"),
        message=b"a simple revision with no parents this time",
        author=Person(
            name=b"Roberto Dicosmo",
            email=b"roberto@example.com",
            fullname=b"Roberto Dicosmo <roberto@example.com>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"tony",
            email=b"ar@dumont.fr",
            fullname=b"tony <ar@dumont.fr>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1127351742,
                microseconds=220000,
            ),
            offset_bytes=b"+0000",
        ),
        parents=tuple([revision.id, revision2.id]),
        type=RevisionType.GIT,
        directory=directory2.id,
        metadata=None,
        extra_headers=(),
        synthetic=True,
    )
    revision4 = Revision(
        id=hash_to_bytes("ae860aec43700c7f5a295e2ef47e2ae41b535dfe"),
        message=b"parent of self.revision2",
        author=Person(
            name=b"me",
            email=b"me@soft.heri",
            fullname=b"me <me@soft.heri>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"committer-dude",
            email=b"committer@dude.com",
            fullname=b"committer-dude <committer@dude.com>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1244567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        parents=tuple([revision3.id]),
        type=RevisionType.GIT,
        directory=directory.id,
        metadata=None,
        extra_headers=(),
        synthetic=False,
    )
    git_revisions: Tuple[Revision, ...] = (revision, revision2, revision3, revision4)

    hg_revision = Revision(
        id=hash_to_bytes("951c9503541e7beaf002d7aebf2abd1629084c68"),
        message=b"hello",
        author=Person(
            name=b"Nicolas Dandrimont",
            email=b"nicolas@example.com",
            fullname=b"Nicolas Dandrimont <nicolas@example.com> ",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1234567890, microseconds=0),
            offset_bytes=b"+0200",
        ),
        committer=Person(
            name=b"St\xc3fano Zacchiroli",
            email=b"stefano@example.com",
            fullname=b"St\xc3fano Zacchiroli <stefano@example.com>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1123456789, microseconds=0),
            offset_bytes=b"+0200",
        ),
        parents=(),
        type=RevisionType.MERCURIAL,
        directory=directory.id,
        metadata={
            "checksums": {
                "sha1": "tarball-sha1",
                "sha256": "tarball-sha256",
            },
            "signed-off-by": "some-dude",
            "node": "a316dfb434af2b451c1f393496b7eaeda343f543",
        },
        extra_headers=(),
        synthetic=True,
    )
    hg_revision2 = Revision(
        id=hash_to_bytes("df4afb063236300eb13b96a0d7fff03f7b7cbbaf"),
        message=b"hello again",
        author=Person(
            name=b"Roberto Dicosmo",
            email=b"roberto@example.com",
            fullname=b"Roberto Dicosmo <roberto@example.com>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"tony",
            email=b"ar@dumont.fr",
            fullname=b"tony <ar@dumont.fr>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1123456789,
                microseconds=220000,
            ),
            offset_bytes=b"+0000",
        ),
        parents=tuple([hg_revision.id]),
        type=RevisionType.MERCURIAL,
        directory=directory2.id,
        metadata=None,
        extra_headers=(
            (b"node", hash_to_bytes("fa1b7c84a9b40605b67653700f268349a6d6aca1")),
        ),
        synthetic=False,
    )
    hg_revision3 = Revision(
        id=hash_to_bytes("84d8e7081b47ebb88cad9fa1f25de5f330872a37"),
        message=b"a simple revision with no parents this time",
        author=Person(
            name=b"Roberto Dicosmo",
            email=b"roberto@example.com",
            fullname=b"Roberto Dicosmo <roberto@example.com>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"tony",
            email=b"ar@dumont.fr",
            fullname=b"tony <ar@dumont.fr>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1127351742,
                microseconds=220000,
            ),
            offset_bytes=b"+0000",
        ),
        parents=tuple([hg_revision.id, hg_revision2.id]),
        type=RevisionType.MERCURIAL,
        directory=directory2.id,
        metadata=None,
        extra_headers=(
            (b"node", hash_to_bytes("7f294a01c49065a90b3fe8b4ad49f08ce9656ef6")),
        ),
        synthetic=True,
    )
    hg_revision4 = Revision(
        id=hash_to_bytes("4683324ba26dfe941a72cc7552e86eaaf7c27fe3"),
        message=b"parent of self.revision2",
        author=Person(
            name=b"me",
            email=b"me@soft.heri",
            fullname=b"me <me@soft.heri>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1234567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        committer=Person(
            name=b"committer-dude",
            email=b"committer@dude.com",
            fullname=b"committer-dude <committer@dude.com>",
        ),
        committer_date=TimestampWithTimezone(
            timestamp=Timestamp(
                seconds=1244567843,
                microseconds=220000,
            ),
            offset_bytes=b"-1200",
        ),
        parents=tuple([hg_revision3.id]),
        type=RevisionType.MERCURIAL,
        directory=directory.id,
        metadata=None,
        extra_headers=(
            (b"node", hash_to_bytes("f4160af0485c85823d9e829bae2c00b00a2e6297")),
        ),
        synthetic=False,
    )
    hg_revisions: Tuple[Revision, ...] = (
        hg_revision,
        hg_revision2,
        hg_revision3,
        hg_revision4,
    )
    revisions: Tuple[Revision, ...] = git_revisions + hg_revisions

    origins: Tuple[Origin, ...] = (
        Origin(url="https://github.com/user1/repo1"),
        Origin(url="https://github.com/user2/repo1"),
        Origin(url="https://github.com/user3/repo1"),
        Origin(url="https://gitlab.com/user1/repo1"),
        Origin(url="https://gitlab.com/user2/repo1"),
        Origin(url="https://forge.softwareheritage.org/source/repo1"),
        Origin(url="https://example.рф/🏛️.txt"),
    )
    origin, origin2 = origins[:2]

    metadata_authority = MetadataAuthority(
        type=MetadataAuthorityType.DEPOSIT_CLIENT,
        url="http://hal.inria.example.com/",
    )
    metadata_authority2 = MetadataAuthority(
        type=MetadataAuthorityType.REGISTRY,
        url="http://wikidata.example.com/",
    )
    authorities: Tuple[MetadataAuthority, ...] = (
        metadata_authority,
        metadata_authority2,
    )

    metadata_fetcher = MetadataFetcher(
        name="swh-deposit",
        version="0.0.1",
    )
    metadata_fetcher2 = MetadataFetcher(
        name="swh-example",
        version="0.0.1",
    )
    fetchers: Tuple[MetadataFetcher, ...] = (metadata_fetcher, metadata_fetcher2)

    date_visit1 = datetime.datetime(2015, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)
    date_visit2 = datetime.datetime(2017, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)
    date_visit3 = datetime.datetime(2018, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)

    type_visit1 = "git"
    type_visit2 = "hg"
    type_visit3 = "deb"

    origin_visit = OriginVisit(
        origin=origin.url,
        visit=1,
        date=date_visit1,
        type=type_visit1,
    )
    origin_visit2 = OriginVisit(
        origin=origin.url,
        visit=2,
        date=date_visit2,
        type=type_visit1,
    )
    origin_visit3 = OriginVisit(
        origin=origin2.url,
        visit=1,
        date=date_visit1,
        type=type_visit2,
    )
    origin_visits: Tuple[OriginVisit, ...] = (
        origin_visit,
        origin_visit2,
        origin_visit3,
    )

    origin_visit_status = OriginVisitStatus(
        origin=origin.url,
        visit=1,
        date=date_visit1,
        type=type_visit1,
        status="created",
        snapshot=None,
    )
    origin_visit2_status = OriginVisitStatus(
        origin=origin.url,
        visit=2,
        date=date_visit2,
        type=type_visit1,
        status="created",
        snapshot=None,
    )
    origin_visit3_status = OriginVisitStatus(
        origin=origin2.url,
        visit=1,
        date=date_visit1,
        type=type_visit2,
        status="created",
        snapshot=None,
    )
    origin_visit_statuses: Tuple[OriginVisitStatus, ...] = (
        origin_visit_status,
        origin_visit2_status,
        origin_visit3_status,
    )

    release = Release(
        id=hash_to_bytes("f7f222093a18ec60d781070abec4a630c850b837"),
        name=b"v0.0.1",
        author=Person(
            name=b"olasd",
            email=b"nic@olasd.fr",
            fullname=b"olasd <nic@olasd.fr>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1234567890, microseconds=0),
            offset_bytes=b"+0042",
        ),
        target=revision.id,
        target_type=ObjectType.REVISION,
        message=b"synthetic release",
        synthetic=True,
    )
    release2 = Release(
        id=hash_to_bytes("db81a26783a3f4a9db07b4759ffc37621f159bb2"),
        name=b"v0.0.2",
        author=Person(
            name=b"tony",
            email=b"ar@dumont.fr",
            fullname=b"tony <ar@dumont.fr>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1634366813, microseconds=0),
            offset_bytes=b"-0200",
        ),
        target=revision2.id,
        target_type=ObjectType.REVISION,
        message=b"v0.0.2\nMisc performance improvements + bug fixes",
        synthetic=False,
    )
    release3 = Release(
        id=hash_to_bytes("1c5d42e603ce2eea44917fadca76c78bad76aeb9"),
        name=b"v0.0.2",
        author=Person(
            name=b"tony",
            email=b"tony@ardumont.fr",
            fullname=b"tony <tony@ardumont.fr>",
        ),
        date=TimestampWithTimezone(
            timestamp=Timestamp(seconds=1634366813, microseconds=0),
            offset_bytes=b"-0200",
        ),
        target=revision3.id,
        target_type=ObjectType.REVISION,
        message=b"yet another synthetic release",
        synthetic=True,
    )

    releases: Tuple[Release, ...] = (release, release2, release3)

    snapshot = Snapshot(
        id=hash_to_bytes("9b922e6d8d5b803c1582aabe5525b7b91150788e"),
        branches={
            b"master": SnapshotBranch(
                target=revision.id,
                target_type=SnapshotTargetType.REVISION,
            ),
        },
    )
    empty_snapshot = Snapshot(
        id=hash_to_bytes("1a8893e6a86f444e8be8e7bda6cb34fb1735a00e"),
        branches={},
    )
    complete_snapshot = Snapshot(
        id=hash_to_bytes("db99fda25b43dc5cd90625ee4b0744751799c917"),
        branches={
            b"directory": SnapshotBranch(
                target=directory.id,
                target_type=SnapshotTargetType.DIRECTORY,
            ),
            b"directory2": SnapshotBranch(
                target=directory2.id,
                target_type=SnapshotTargetType.DIRECTORY,
            ),
            b"content": SnapshotBranch(
                target=content.sha1_git,
                target_type=SnapshotTargetType.CONTENT,
            ),
            b"alias": SnapshotBranch(
                target=b"revision",
                target_type=SnapshotTargetType.ALIAS,
            ),
            b"revision": SnapshotBranch(
                target=revision.id,
                target_type=SnapshotTargetType.REVISION,
            ),
            b"release": SnapshotBranch(
                target=release.id,
                target_type=SnapshotTargetType.RELEASE,
            ),
            b"snapshot": SnapshotBranch(
                target=empty_snapshot.id,
                target_type=SnapshotTargetType.SNAPSHOT,
            ),
            b"dangling": None,
        },
    )

    snapshots: Tuple[Snapshot, ...] = (snapshot, empty_snapshot, complete_snapshot)

    content_metadata1 = RawExtrinsicMetadata(
        target=ExtendedSWHID(
            object_type=ExtendedObjectType.CONTENT, object_id=content.sha1_git
        ),
        origin=origin.url,
        discovery_date=datetime.datetime(
            2015, 1, 1, 21, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=metadata_authority,
        fetcher=metadata_fetcher,
        format="json",
        metadata=b'{"foo": "bar"}',
    )
    content_metadata2 = RawExtrinsicMetadata(
        target=ExtendedSWHID(
            object_type=ExtendedObjectType.CONTENT, object_id=content.sha1_git
        ),
        origin=origin2.url,
        discovery_date=datetime.datetime(
            2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=metadata_authority,
        fetcher=metadata_fetcher,
        format="yaml",
        metadata=b"foo: bar",
    )
    content_metadata3 = RawExtrinsicMetadata(
        target=ExtendedSWHID(
            object_type=ExtendedObjectType.CONTENT, object_id=content.sha1_git
        ),
        discovery_date=datetime.datetime(
            2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=attr.evolve(metadata_authority2, metadata=None),
        fetcher=attr.evolve(metadata_fetcher2, metadata=None),
        format="yaml",
        metadata=b"foo: bar",
        origin=origin.url,
        visit=42,
        snapshot=snapshot.swhid(),
        release=release.swhid(),
        revision=revision.swhid(),
        directory=directory.swhid(),
        path=b"/foo/bar",
    )

    content_metadata: Tuple[RawExtrinsicMetadata, ...] = (
        content_metadata1,
        content_metadata2,
        content_metadata3,
    )

    origin_metadata1 = RawExtrinsicMetadata(
        target=Origin(origin.url).swhid(),
        discovery_date=datetime.datetime(
            2015, 1, 1, 21, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=attr.evolve(metadata_authority, metadata=None),
        fetcher=attr.evolve(metadata_fetcher, metadata=None),
        format="json",
        metadata=b'{"foo": "bar"}',
    )
    origin_metadata2 = RawExtrinsicMetadata(
        target=Origin(origin.url).swhid(),
        discovery_date=datetime.datetime(
            2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=attr.evolve(metadata_authority, metadata=None),
        fetcher=attr.evolve(metadata_fetcher, metadata=None),
        format="yaml",
        metadata=b"foo: bar",
    )
    origin_metadata3 = RawExtrinsicMetadata(
        target=Origin(origin.url).swhid(),
        discovery_date=datetime.datetime(
            2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
        ),
        authority=attr.evolve(metadata_authority2, metadata=None),
        fetcher=attr.evolve(metadata_fetcher2, metadata=None),
        format="yaml",
        metadata=b"foo: bar",
    )

    origin_metadata: Tuple[RawExtrinsicMetadata, ...] = (
        origin_metadata1,
        origin_metadata2,
        origin_metadata3,
    )

    extid1 = ExtID(
        target=CoreSWHID(object_type=SwhidObjectType.REVISION, object_id=revision.id),
        extid_type="git",
        extid=revision.id,
    )

    extid2 = ExtID(
        target=CoreSWHID(
            object_type=SwhidObjectType.REVISION, object_id=hg_revision.id
        ),
        extid_type="mercurial",
        extid=hash_to_bytes("a316dfb434af2b451c1f393496b7eaeda343f543"),
    )

    extid3 = ExtID(
        target=CoreSWHID(object_type=SwhidObjectType.DIRECTORY, object_id=directory.id),
        extid_type="directory",
        extid=b"something",
    )
    extid4 = ExtID(
        target=CoreSWHID(
            object_type=SwhidObjectType.DIRECTORY, object_id=directory2.id
        ),
        extid_type="directory",
        extid=b"something",
        extid_version=2,
    )

    extids: Tuple[ExtID, ...] = (
        extid1,
        extid2,
        extid3,
        extid4,
    )
